# -*- coding: utf-8 -*-
#
# Copyright (C) 2005-2006 Edgewall Software
# Copyright (C) 2005-2006 Matthew Good <trac@matt-good.net>
# Copyright (C) 2006 Christopher Lenz <cmlenz@gmx.de>
# All rights reserved.
#
# This software is licensed as described in the file COPYING, which
# you should have received as part of this distribution. The terms
# are also available at http://trac.edgewall.com/license.html.
#
# This software consists of voluntary contributions made by many
# individuals. For the exact contribution history, see the revision
# history and logs, available at http://projects.edgewall.com/trac/.
#
# Author: Matthew Good <trac@matt-good.net>
#         Christopher Lenz <cmlenz@gmx.de>

import urllib2
from email.utils import parseaddr

from trac.config import IntOption, Option
from trac.core import Component, implements
from trac.mimeview.api import is_binary
from trac.util.text import unicode_urlencode

from tracspamfilter.api import IFilterStrategy, N_, user_agent


class AkismetFilterStrategy(Component):
    """Spam filter using the Akismet service (http://akismet.com/).

    Based on the `akismet` Python module written by Michael Ford:
      http://www.voidspace.org.uk/python/modules.shtml#akismet
    """
    implements(IFilterStrategy)

    noheaders = ['HTTP_COOKIE', 'HTTP_HOST', 'HTTP_REFERER',
                 'HTTP_USER_AGENT', 'HTTP_AUTHORIZATION']

    karma_points = IntOption('spam-filter', 'akismet_karma', '10',
        """By how many points an Akismet reject impacts the overall karma of
        a submission.""", doc_domain='tracspamfilter')

    api_key = Option('spam-filter', 'akismet_api_key', '',
        """Wordpress key required to use the Akismet API.""",
        doc_domain='tracspamfilter')

    api_url = Option('spam-filter', 'akismet_api_url',
        'rest.akismet.com/1.1/', """URL of the Akismet service.""",
        doc_domain='tracspamfilter')

    def __init__(self):
        self.verified_key = None
        self.verified = None

    # IFilterStrategy methods

    def is_external(self):
        return True

    def test(self, req, author, content, ip):
        if not self._check_preconditions(req, author, content):
            return

        try:
            url = 'http://%s.%scomment-check' % (self.api_key, self.api_url)
            self.log.debug("Checking content with Akismet service at %s", url)
            resp = self._post(url, req, author, content, ip)
            if resp.strip().lower() != 'false':
                self.log.debug("Akismet says content is spam")
                return -abs(self.karma_points), \
                       N_("Akismet says content is spam")

        except urllib2.URLError, e:
            self.log.warn("Akismet request failed (%s)", e)

    def train(self, req, author, content, ip, spam=True):
        if not self._check_preconditions(req, author, content):
            return -2

        try:
            which = spam and 'spam' or 'ham'
            url = 'http://%s.%ssubmit-%s' \
                  % (self.api_key, self.api_url, which)
            self.log.debug("Submitting %s to Akismet service at %s",
                           which, url)
            self._post(url, req, author, content, ip)
            return 1
        except urllib2.URLError, e:
            self.log.warn("Akismet request failed (%s)", e)
        return -1

    # Internal methods

    def _check_preconditions(self, req, author, content):
        if self.karma_points == 0:
            return False

        if not self.api_key:
            self.log.warning("Akismet API key is missing")
            return False

        if is_binary(content):
            self.log.warning("Content is binary, Akismet content check "
                             "skipped")
            return False

        try:
            if not self.verify_key(req):
                self.log.warning("Akismet API key is invalid")
                return False
            return True
        except urllib2.URLError, e:
            self.log.warn("Akismet request failed (%s)", e)

    def verify_key(self, req, api_url=None, api_key=None):
        if api_url is None:
            api_url = self.api_url
        if api_key is None:
            api_key = self.api_key

        if api_key != self.verified_key:
            self.log.debug("Verifying Akismet API key")
            params = {'blog': req.base_url, 'key': api_key}
            req = urllib2.Request('http://%sverify-key' % api_url,
                                  unicode_urlencode(params),
                                  {'User-Agent': user_agent})
            resp = urllib2.urlopen(req).read()
            if resp.strip().lower() == 'valid':
                self.log.debug("Akismet API key is valid")
                self.verified = True
                self.verified_key = api_key

        return self.verified_key is not None

    def _post(self, url, req, author, content, ip):
        # Split up author into name and email, if possible
        author_name, author_email = parseaddr(author)
        if not author_name and not author_email:
            author_name = author
        elif not author_name and author_email.find('@') < 1:
            author_name = author
            author_email = None

        params = {
            'blog': req.base_url,
            'user_ip': ip,
            'user_agent': req.get_header('User-Agent'),
            'referrer': req.get_header('Referer') or 'unknown',
            'comment_author': author_name,
            'comment_type': 'trac',
            'comment_content': content
        }
        if author_email:
            params['comment_author_email'] = author_email
        for k, v in req.environ.items():
            if k.startswith('HTTP_') and k not in self.noheaders:
                params[k] = v
        urlreq = urllib2.Request(url, unicode_urlencode(params),
                                 {'User-Agent': user_agent})

        resp = urllib2.urlopen(urlreq)
        return resp.read()



# -*- coding: utf-8 -*-
#
# Copyright (C) 2015 Edgewall Software
# Copyright (C) 2006 Matthew Good <trac@matt-good.net>
# Copyright (C) 2015 Dirk Stöcker <trac@dstoecker.de>
# All rights reserved.
#
# This software is licensed as described in the file COPYING, which
# you should have received as part of this distribution. The terms
# are also available at http://trac.edgewall.com/license.html.
#
# This software consists of voluntary contributions made by many
# individuals. For the exact contribution history, see the revision
# history and logs, available at http://projects.edgewall.com/trac/.
#
# Author: Matthew Good <trac@matt-good.net>

from dns.ipv6 import inet_aton as ipv6_inet_aton
from dns.name import from_text
from dns.resolver import NXDOMAIN, NoAnswer, NoNameservers, Timeout, query

from trac.config import ListOption, IntOption
from trac.core import Component, implements

from tracspamfilter.api import IFilterStrategy, N_


class IPBlacklistFilterStrategy(Component):
    """Spam filter based on IP blacklistings.

    Requires the dnspython module from http://www.dnspython.org/.
    """
    implements(IFilterStrategy)

    karma_points = IntOption('spam-filter', 'ip_blacklist_karma', '5',
        """By how many points blacklisting by a single server impacts the
        overall karma of a submission.""", doc_domain='tracspamfilter')

    servers_default = 'all.s5h.net,dnsbl.dronebl.org,rbl.blockedservers.com'
    servers = ListOption('spam-filter', 'ip_blacklist_servers',
        servers_default, doc="Servers used for IPv4 blacklisting.",
        doc_domain='tracspamfilter')

    servers6_default = 'all.s5h.net,dnsbl.dronebl.org,' \
                       'bl.ipv6.spameatingmonkey.net'
    servers6 = ListOption('spam-filter', 'ip6_blacklist_servers',
        servers6_default, doc="Servers used for IPv6 blacklisting.",
        doc_domain='tracspamfilter')

    # IFilterStrategy implementation

    def is_external(self):
        return True

    def test(self, req, author, content, ip):
        if self.karma_points == 0:
            return

        serverlist, prefix = self._getdata(ip)
        if not serverlist:
            self.log.warning("No IP blacklist servers configured")
            return

        self.log.debug('Checking for IP blacklisting on "%s"', ip)

        points = 0
        servers = []

        for server in serverlist:
            self.log.debug("Checking blacklist %s for %s [%s]", server, ip,
                           prefix)
            try:
                res = query(from_text(prefix +
                                      server.encode('utf-8')))[0].to_text()
                points -= abs(self.karma_points)
                if res == '127.0.0.1':
                    servers.append(server)
                else:
                    # strip the common part of responses
                    if res.startswith('127.0.0.'):
                        res = res[8:]
                    elif res.startswith('127.'):
                        res = res[4:]
                    servers.append('%s [%s]' % (server, res))
            except NXDOMAIN:  # not blacklisted on this server
                continue
            except (Timeout, NoAnswer, NoNameservers), e:
                self.log.warning('Error checking IP blacklist server "%s" '
                                 'for IP "%s": %s', server, ip, e)

        if points != 0:
            return (points, N_("IP %s blacklisted by %s"), ip,
                    ', '.join(servers))

    def train(self, req, author, content, ip, spam=True):
        return 0

    # Internal methods

    def _getdata(self, ip):
        if ip.find(".") < 0:
            encoded = reversed(list(ipv6_inet_aton(ip).encode('hex_codec')))
            return self.servers6, '.'.join(encoded) + '.'
        return self.servers, '.'.join(reversed(ip.split('.'))) + '.'



# -*- coding: utf-8 -*-
#
# Copyright (C) 2006 Edgewall Software
# All rights reserved.
#
# This software is licensed as described in the file COPYING, which
# you should have received as part of this distribution. The terms
# are also available at http://trac.edgewall.com/license.html.
#
# This software consists of voluntary contributions made by many
# individuals. For the exact contribution history, see the revision
# history and logs, available at http://projects.edgewall.com/trac/.

from acct_mgr.api import IAccountRegistrationInspector
from acct_mgr.register import RegistrationError

from trac.config import BoolOption, IntOption
from trac.core import Component, ExtensionPoint, implements
from trac.util.html import tag

from tracspamfilter.api import IFilterStrategy, N_


class RegistrationFilterStrategy(Component):
    """Spam filter strategy that calls account manager checks for
    account registration.
    """
    implements(IFilterStrategy)

    karma_points = IntOption('spam-filter', 'account_karma', '0',
        """By how many points a failed registration check impacts
        the overall score.""", doc_domain='tracspamfilter')

    replace_checks = BoolOption('spam-filter', 'account_replace_checks',
        'false',
        """Replace checks in account manager totally.""",
        doc_domain='tracspamfilter')

    listeners = ExtensionPoint(IAccountRegistrationInspector)

    def render_registration_fields(self, req, data, fragments):
        self.log.debug("Adding registration check data fields")
        if self.replace_checks:
            for check in self.listeners:
                try:
                    if check.__class__.__name__ != 'RegistrationFilterAdapter':
                        self.log.debug("Add registration check data %s", check)
                        fragment, f_data = \
                            check.render_registration_fields(req, data)
                        try:
                            fragments['optional'] = \
                                tag(fragments.get('optional', ''),
                                    fragment.get('optional', ''))
                            fragments['required'] = \
                                tag(fragments.get('required', ''),
                                    fragment.get('required', ''))
                        except AttributeError:
                            if fragment is not None and fragment != '':
                                fragments['required'] = \
                                    tag(fragments.get('required', ''),
                                        fragment)
                        data.update(f_data)
                except Exception, e:
                    self.log.exception("Adding registration fields failed: %s",
                                       e)
        return fragments, data

    # IFilterStrategy methods

    def is_external(self):
        return False

    def test(self, req, author, content, ip):
        if req.path_info == '/register':
            karma = 0
            checks = []
            for check in self.listeners:
                try:
                    if check.__class__.__name__ != 'RegistrationFilterAdapter':
                        self.log.debug("Try registration check %s", check)
                        check.validate_registration(req)
                except RegistrationError, e:
                    karma -= abs(self.karma_points)
                    msg = e.message.replace('\n', '')
                    args = e.msg_args
                    if args:
                        msg = msg % args
                        msg.replace('<b>', '*').replace('</b>', '*')
                    self.log.debug("Registration check returned %s", msg)
                    checks.append('%s: %s' % (check.__class__.__name__, msg))
                except Exception, e:
                    self.log.exception("Registration check %s failed: %s",
                                       check, e)
            if karma or checks:
                return karma, N_("Account registration failed (%s)"), \
                       ", ".join(checks)

    def train(self, req, author, content, ip, spam=True):
        return 0



# -*- coding: utf-8 -*-
#
# Copyright (C) 2005-2006 Edgewall Software
# Copyright (C) 2005 Matthew Good <trac@matt-good.net>
# Copyright (C) 2006 Christopher Lenz <cmlenz@gmx.de>
# All rights reserved.
#
# This software is licensed as described in the file COPYING, which
# you should have received as part of this distribution. The terms
# are also available at http://trac.edgewall.com/license.html.
#
# This software consists of voluntary contributions made by many
# individuals. For the exact contribution history, see the revision
# history and logs, available at http://projects.edgewall.com/trac/.
#
# Author: Matthew Good <trac@matt-good.net>

import re

from trac.config import BoolOption, IntOption, Option
from trac.core import Component, TracError, implements
from trac.util.html import tag
from trac.wiki.api import IWikiChangeListener, IWikiPageManipulator
from trac.wiki.model import WikiPage

from tracspamfilter.api import IFilterStrategy, N_, tag_


class RegexFilterStrategy(Component):
    """Spam filter based on regular expressions defined in BadContent page.
    """
    implements(IFilterStrategy, IWikiChangeListener, IWikiPageManipulator)

    karma_points = IntOption('spam-filter', 'regex_karma', '5',
        """By how many points a match with a pattern on the BadContent page
        impacts the overall karma of a submission.""",
        doc_domain='tracspamfilter')

    badcontent_file = Option('spam-filter', 'badcontent_file', '',
        """Local file to be loaded to get BadContent. Can be used in
        addition to BadContent wiki page.""",
        doc_domain='tracspamfilter')

    show_blacklisted = BoolOption('spam-filter', 'show_blacklisted', 'true',
        "Show the matched bad content patterns in rejection message.",
        doc_domain='tracspamfilter')

    def __init__(self):
        self.patterns = []
        page = WikiPage(self.env, 'BadContent')
        if page.exists:
            try:
                self._load_patterns(page)
            except TracError:
                pass
        if self.badcontent_file != '':
            with open(self.badcontent_file, 'r') as file:
                if file is None:
                    self.log.warning("BadContent file cannot be opened")
                else:
                    lines = file.read().splitlines()
                    pat = [re.compile(p.strip()) for p in lines if p.strip()]
                    self.log.debug("Loaded %s patterns from BadContent file",
                                   len(pat))
                    self.patterns += pat

    # IFilterStrategy implementation

    def is_external(self):
        return False

    def test(self, req, author, content, ip):
        gotcha = []
        points = 0
        if author is not None and author != 'anonymous':
            testcontent = author + '\n' + content
        else:
            testcontent = content
        for pattern in self.patterns:
            match = pattern.search(testcontent)
            if match:
                gotcha.append("'%s'" % pattern.pattern)
                self.log.debug('Pattern %s found in submission',
                               pattern.pattern)
                points -= abs(self.karma_points)
        if points != 0:
            if self.show_blacklisted:
                matches = ", ".join(gotcha)
                return points, N_("Content contained these blacklisted "
                                  "patterns: %s"), matches
            else:
                return points, N_("Content contained %s blacklisted "
                                  "patterns"), str(len(gotcha))

    def train(self, req, author, content, ip, spam=True):
        return 0

    # IWikiPageManipulator implementation

    def prepare_wiki_page(self, req, page, fields):
        pass

    def validate_wiki_page(self, req, page):
        if page.name == 'BadContent':
            try:
                self._load_patterns(page)
            except TracError, e:
                return [(None, e)]
        return []

    # IWikiChangeListener implementation

    def wiki_page_changed(self, page, *args):
        if page.name == 'BadContent':
            self._load_patterns(page)

    wiki_page_added = wiki_page_changed

    wiki_page_version_deleted = wiki_page_changed

    def wiki_page_deleted(self, page):
        if page.name == 'BadContent':
            self.patterns = []

    # Internal methods

    def _load_patterns(self, page):
        if '{{{' in page.text and '}}}' in page.text:
            lines = page.text.split('{{{', 1)[1].split('}}}', 1)[0]
            lines = [l.strip() for l in lines.splitlines() if l.strip()]
            for p in lines:
                try:
                    self.patterns.append(re.compile(p))
                except re.error, e:
                    self.log.debug("Error in pattern %s: %s", p, e)
                    raise TracError(tag_("Error in pattern %(pattern)s: "
                                         "%(error)s.", pattern=tag.tt(p),
                                         error=tag.i(e)))
            self.log.debug("Loaded %s patterns from BadContent",
                           len(self.patterns))
        else:
            self.log.warning("BadContent page does not contain any patterns")
            self.patterns = []



# -*- coding: utf-8 -*-
#
# Copyright (C) 2006 Edgewall Software
# All rights reserved.
#
# This software is licensed as described in the file COPYING, which
# you should have received as part of this distribution. The terms
# are also available at http://trac.edgewall.com/license.html.
#
# This software consists of voluntary contributions made by many
# individuals. For the exact contribution history, see the revision
# history and logs, available at http://projects.edgewall.com/trac/.

import time
import unittest
from datetime import timedelta

from trac.db.api import DatabaseManager
from trac.test import EnvironmentStub
from trac.util.datefmt import datetime_now, to_utimestamp, utc

from tracspamfilter.filtersystem import FilterSystem
from tracspamfilter.model import LogEntry


def reset_db(env):
    DatabaseManager(env).drop_tables(
        ('spamfilter_bayes', 'spamfilter_log',
         'spamfilter_report', 'spamfilter_statistics'))
    env.reset_db()


class LogEntryTestCase(unittest.TestCase):

    def setUp(self):
        self.env = EnvironmentStub()
        FilterSystem(self.env).upgrade_environment()

    def tearDown(self):
        reset_db(self.env)

    def _insert_log_entries(self, num_entries):
        """Helper method to insert log entries."""
        now = to_utimestamp(datetime_now(utc))
        self.env.db_transaction.executemany("""
            INSERT INTO spamfilter_log
             (time,path,author,authenticated,ipnr,headers,content,
              rejected,karma,reasons,request)
            VALUES (%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s)
            """, [(now + i, '/ticket/1', 'Joe Smith <joesmith@example.com',
                   0, '::1', '', "Test", 1, -1, None, None)
                  for i in xrange(0, num_entries)])

    def test_delete_one(self):
        """Delete one entry."""
        n = 1
        self._insert_log_entries(n)

        self.assertEqual(n, len(list(LogEntry.select(self.env))))
        LogEntry.delete(self.env, n)
        self.assertEqual(0, len(list(LogEntry.select(self.env))))

    def test_delete_many(self):
        """Delete many entries."""
        n = 10
        self._insert_log_entries(n)

        self.assertEqual(n, len(list(LogEntry.select(self.env))))
        LogEntry.delete(self.env, range(1, n + 1))
        self.assertEqual(0, len(list(LogEntry.select(self.env))))

    def test_purge(self):
        now = datetime_now()
        oneweekago = time.mktime((now - timedelta(weeks=1)).timetuple())
        onedayago = time.mktime((now - timedelta(days=1)).timetuple())
        req = None

        LogEntry(self.env, oneweekago, '/foo', 'john', False, '127.0.0.1',
                 '', 'Test', False, 5, [], req).insert()
        LogEntry(self.env, onedayago, '/foo', 'anonymous', False, '127.0.0.1',
                 '', 'Test', True, -3, [], req).insert()

        LogEntry.purge(self.env, days=4)

        log = list(LogEntry.select(self.env))
        self.assertEqual(1, len(log))
        entry = log[0]
        self.assertEqual('anonymous', entry.author)


def test_suite():
    suite = unittest.TestSuite()
    suite.addTest(unittest.makeSuite(LogEntryTestCase))
    return suite

if __name__ == '__main__':
    unittest.main(defaultTest='test_suite')



# -*- coding: utf-8 -*-
#
# Copyright (C) 2014-2016 Edgewall Software
# All rights reserved.
#
# This software is licensed as described in the file COPYING, which
# you should have received as part of this distribution. The terms
# are also available at http://trac.edgewall.com/license.html.
#
# This software consists of voluntary contributions made by many
# individuals. For the exact contribution history, see the revision
# history and logs, available at http://projects.edgewall.com/trac/.

from time import time

from trac.config import ListOption
from trac.core import Component, TracError, implements
from trac.util.html import tag
from trac.web.api import IRequestFilter, IRequestHandler
from trac.web.chrome import (
    add_ctxtnav, add_notice, add_script, add_script_data)

from tracspamfilter.api import _, ngettext


class SpamReportAdapter(Component):
    """Interface to allow users to report spam."""

    checkreq = ListOption('spam-filter', 'report_pages',
                                 'wiki, attachment, ticket', doc=
        """List of page types to add spam report link""",
        doc_domain="tracspamfilter")

    implements(IRequestFilter, IRequestHandler)

    def match_request(self, req):
        return req.path_info == '/reportspam'

    def process_request(self, req):
        req.perm.require('SPAM_REPORT')
        if 'page' not in req.args:
            raise TracError(_("No page supplied to report as spam"))
        page = req.args['page']
        savepage = page
        isauth = 1 if req.authname and req.authname != 'anonymous' else 0
        headers = '\n'.join(['%s: %s' % (k[5:].replace('_', '-').title(), v)
                             for k, v in req.environ.items()
                             if k.startswith('HTTP_')])
        if page.startswith("/ticket/"):
            for tim, in self.env.db_query("SELECT time/1000000 from ticket WHERE id=%s", (page[8:],)):
                # append creation time, so spam log entries can be seen after deleting ticket
                savepage = "%s#%d" % (page, tim)
        self.env.db_transaction("""
            INSERT INTO spamfilter_report
                (entry, headers, author, authenticated, comment, time)
                VALUES (%s,%s,%s,%s,%s,%s)""",
            (savepage, headers, req.authname, isauth, req.args.get('comment', None), int(time())))
        req.redirect(req.href(page))

    def pre_process_request(self, req, handler):
        return handler

    def post_process_request(self, req, template, data, content_type):
        if 'SPAM_REPORT' in req.perm:
            i = req.path_info.find("/", 1)
            if (req.path_info[1:i] if i > 0 else req.path_info[1:]) in self.checkreq:
                isauth = 1 if req.authname and req.authname != 'anonymous' else 0
                if self.env.db_query("""
                    SELECT id FROM spamfilter_report
                       WHERE entry = %s
                          AND author = %s
                          AND authenticated = %s""",
                                     (req.path_info, req.authname, isauth)):
                    add_ctxtnav(req, _('Reported spam'))
                else:
                    add_script_data(req, {'spamreport_comment': _("Comment")})
                    add_script(req, 'spamfilter/reportspam.js')
                    add_ctxtnav(req, tag.a(_('Report spam'), id='reportspam', href=req.href('reportspam', (('page', req.path_info),))))
        if 'SPAM_CHECKREPORTS' in req.perm:
            for total, in self.env.db_query("SELECT COUNT(*) FROM spamfilter_report"):
                if total:
                    add_notice(req, tag.a(ngettext('%(num)d spam report', '%(num)d spam reports', total),
                    href=req.href.admin("spamfilter/report")))

        return template, data, content_type



# -*- coding: utf-8 -*-
#
# Copyright (C) 2015-2016 Edgewall Software
# Copyright (C) 2015 Dirk Stöcker <trac@dstoecker.de>
# All rights reserved.
#
# This software is licensed as described in the file COPYING, which
# you should have received as part of this distribution. The terms
# are also available at http://trac.edgewall.com/license.html.
#
# This software consists of voluntary contributions made by many
# individuals. For the exact contribution history, see the revision
# history and logs, available at http://projects.edgewall.com/trac/.
#
# Author: Dirk Stöcker <trac@dstoecker.de>

from trac.admin import IAdminPanelProvider
from trac.core import Component, ExtensionPoint, implements
from trac.util.html import tag
from trac.util.translation import tag_
from trac.web.chrome import add_stylesheet, add_warning

from tracspamfilter.api import _
from tracspamfilter.captcha import ICaptchaMethod
from tracspamfilter.captcha.expression import ExpressionCaptcha
from tracspamfilter.captcha.keycaptcha import KeycaptchaCaptcha
from tracspamfilter.captcha.recaptcha2 import Recaptcha2Captcha

try:
    from tracspamfilter.captcha.image import ImageCaptcha
except ImportError:  # PIL not installed
    ImageCaptcha = None


class CaptchaAdminPageProvider(Component):
    """Web administration panel for configuring the Captcha handling."""

    implements(IAdminPanelProvider)

    handlers = ExtensionPoint(ICaptchaMethod)

    def __init__(self):
        self.recaptcha_enabled = \
            self.env.is_enabled(Recaptcha2Captcha)
        self.keycaptcha_enabled =\
            self.env.is_enabled(KeycaptchaCaptcha)
        self.expressioncaptcha_enabled = \
            self.env.is_enabled(ExpressionCaptcha)
        self.imagecaptcha_enabled = \
            ImageCaptcha and self.env.is_enabled(ImageCaptcha)

    # IAdminPanelProvider methods

    def get_admin_panels(self, req):
        any_captchas_enabled = \
            any((self.recaptcha_enabled, self.keycaptcha_enabled,
                 self.expressioncaptcha_enabled, self.imagecaptcha_enabled))
        if 'SPAM_CONFIG' in req.perm and any_captchas_enabled:
            yield 'spamfilter', _("Spam Filtering"), 'captcha', _("Captcha")

    def render_admin_panel(self, req, cat, page, path_info):
        req.perm.require('SPAM_CONFIG')

        spam_config = self.config['spam-filter']
        data = {}

        if req.method == 'POST':
            if 'cancel' in req.args:
                req.redirect(req.href.admin(cat, page))

            captcha_enabled = 'captcha_enabled' in req.args
            captcha = req.args.get('captcha')

            captcha_karma_lifetime = req.args.as_int('captcha_karma_lifetime')
            if captcha_karma_lifetime is None:
                captcha_karma_lifetime = \
                    spam_config.get('captcha_karma_lifetime')
                add_warning(req,
                            tag_("Invalid value for %(key)s",
                                 key=tag.tt('captcha_karma_lifetime')))

            recaptcha_private_key = recaptcha_public_key = None
            if self.recaptcha_enabled:
                recaptcha_private_key = req.args.get('recaptcha_private_key')
                recaptcha_public_key = req.args.get('recaptcha_public_key')
                recaptcha = Recaptcha2Captcha(self.env)
                verified_key = recaptcha.verify_key(recaptcha_private_key,
                                                    recaptcha_public_key)
                if recaptcha_private_key and not verified_key:
                    data['recaptcha_error'] = _("The keys are invalid")
                    data['error'] = 1

            keycaptcha_private_key = keycaptcha_user_id = None
            if self.keycaptcha_enabled:
                keycaptcha_private_key = req.args.get('keycaptcha_private_key')
                keycaptcha_user_id = req.args.get('keycaptcha_user_id')
                keycaptcha = KeycaptchaCaptcha(self.env)
                verified_key = keycaptcha.verify_key(keycaptcha_private_key,
                                                     keycaptcha_user_id)
                if keycaptcha_private_key and not verified_key:
                    data['keycaptcha_error'] = \
                        _("The key or user id are invalid")
                    data['error'] = 1

            expressioncaptcha_ceiling = expressioncaptcha_terms = None
            if self.expressioncaptcha_enabled:
                expressioncaptcha_ceiling = \
                    req.args.as_int('expressioncaptcha_ceiling')
                if expressioncaptcha_ceiling is None:
                    expressioncaptcha_ceiling = \
                        spam_config.get('captcha_expression_ceiling')
                    add_warning(req,
                                tag_("Invalid value for %(key)s",
                                     key=tag.tt('captcha_expression_ceiling')))

                expressioncaptcha_terms = \
                    req.args.as_int('expressioncaptcha_terms')
                if expressioncaptcha_terms is None:
                    expressioncaptcha_terms = \
                        spam_config.get('captcha_expression_terms')
                    add_warning(req,
                                tag_("Invalid value for %(key)s",
                                     key=tag.tt('captcha_expression_terms')))

            imagecaptcha_alphabet = imagecaptcha_fonts = \
                imagecaptcha_letters = imagecaptcha_font_size = None
            if self.imagecaptcha_enabled:
                imagecaptcha_alphabet = req.args.get('imagecaptcha_alphabet')
                imagecaptcha_fonts = req.args.get('imagecaptcha_fonts')
                imagecaptcha_letters = req.args.as_int('imagecaptcha_letters')
                if imagecaptcha_letters is None:
                    imagecaptcha_letters = \
                        spam_config.get('captcha_image_letters')
                    add_warning(req,
                                tag_("Invalid value for %(key)s",
                                     key=tag.tt('captcha_image_letters')))

                imagecaptcha_font_size = \
                    req.args.as_int('imagecaptcha_font_size')
                if imagecaptcha_font_size is None:
                    imagecaptcha_font_size = \
                        spam_config.get('captcha_image_font_size')
                    add_warning(req,
                                tag_("Invalid value for %(key)s",
                                     key=tag.tt('captcha_image_font_size')))

            if 'error' not in data or not data['error']:
                spam_config.set('captcha', captcha)
                if captcha_enabled:
                    spam_config.set('reject_handler', 'CaptchaSystem')
                else:
                    spam_config.set('reject_handler', 'FilterSystem')

                spam_config.set('captcha_karma_lifetime',
                                captcha_karma_lifetime)

                if self.recaptcha_enabled:
                    spam_config.set('captcha_recaptcha_private_key',
                                    recaptcha_private_key)
                    spam_config.set('captcha_recaptcha_public_key',
                                    recaptcha_public_key)

                if self.keycaptcha_enabled:
                    spam_config.set('captcha_keycaptcha_private_key',
                                    keycaptcha_private_key)
                    spam_config.set('captcha_keycaptcha_user_id',
                                    keycaptcha_user_id)

                if self.expressioncaptcha_enabled:
                    spam_config.set('captcha_expression_ceiling',
                                    expressioncaptcha_ceiling)
                    spam_config.set('captcha_expression_terms',
                                    expressioncaptcha_terms)

                if self.imagecaptcha_enabled:
                    spam_config.set('captcha_image_alphabet',
                                    imagecaptcha_alphabet)
                    spam_config.set('captcha_image_letters',
                                    imagecaptcha_letters)
                    spam_config.set('captcha_image_font_size',
                                    imagecaptcha_font_size)
                    spam_config.set('captcha_image_fonts',
                                    imagecaptcha_fonts)

                self.config.save()
                req.redirect(req.href.admin(cat, page))

        else:
            captcha = spam_config.get('captcha')
            reject_handler = spam_config.get('reject_handler')
            captcha_enabled = reject_handler == 'CaptchaSystem'
            captcha_karma_lifetime = spam_config.get('captcha_karma_lifetime')

            recaptcha_private_key = \
                spam_config.get('captcha_recaptcha_private_key')
            recaptcha_public_key = \
                spam_config.get('captcha_recaptcha_public_key')

            keycaptcha_private_key = \
                spam_config.get('captcha_keycaptcha_private_key')
            keycaptcha_user_id = \
                spam_config.get('captcha_keycaptcha_user_id')

            expressioncaptcha_ceiling = \
                spam_config.get('captcha_expression_ceiling')
            expressioncaptcha_terms = \
                spam_config.get('captcha_expression_terms')

            imagecaptcha_alphabet = \
                spam_config.get('captcha_image_alphabet')
            imagecaptcha_letters = \
                spam_config.get('captcha_image_letters')
            imagecaptcha_font_size = \
                spam_config.get('captcha_image_font_size')
            imagecaptcha_fonts = \
                spam_config.get('captcha_image_fonts')

        captcha_types = sorted(h.__class__.__name__ for h in self.handlers)

        data.update({
            'captcha': captcha,
            'captcha_types': captcha_types,
            'captcha_enabled': captcha_enabled,
            'captcha_karma_lifetime': captcha_karma_lifetime,
            'recaptcha_enabled': self.recaptcha_enabled,
            'recaptcha_private_key': recaptcha_private_key,
            'recaptcha_public_key': recaptcha_public_key,
            'keycaptcha_enabled': self.keycaptcha_enabled,
            'keycaptcha_private_key': keycaptcha_private_key,
            'keycaptcha_user_id': keycaptcha_user_id,
            'expressioncaptcha_enabled': self.expressioncaptcha_enabled,
            'expressioncaptcha_ceiling': expressioncaptcha_ceiling,
            'expressioncaptcha_terms': expressioncaptcha_terms,
            'imagecaptcha_enabled': self.imagecaptcha_enabled,
            'imagecaptcha_alphabet': imagecaptcha_alphabet,
            'imagecaptcha_letters': imagecaptcha_letters,
            'imagecaptcha_font_size': imagecaptcha_font_size,
            'imagecaptcha_fonts': imagecaptcha_fonts
        })

        add_stylesheet(req, 'spamfilter/admin.css')
        return 'admin_captcha.html', data



# -*- coding: utf-8 -*-
#
# Copyright (C) 2006-2016 Edgewall Software
# All rights reserved.
#
# This software is licensed as described in the file COPYING, which
# you should have received as part of this distribution. The terms
# are also available at https://trac.edgewall.com/license.html.
#
# This software consists of voluntary contributions made by many
# individuals. For the exact contribution history, see the revision
# history and logs, available at http://projects.edgewall.com/trac/.

from trac.db import Column, DatabaseManager, Table


def do_upgrade(env, ver, cursor):
    """Add a column to the log table for storing the complete request.
    Add table required for spam reports by users.
    Add table required for filtering statistics.
    """
    changed_tables = [
        Table('spamfilter_log', key='id')[
            Column('id', auto_increment=True),
            Column('time', type='int'),
            Column('path'),
            Column('author'),
            Column('authenticated', type='int'),
            Column('ipnr'),
            Column('headers'),
            Column('content'),
            Column('rejected', type='int'),
            Column('karma', type='int'),
            Column('reasons'),
            Column('request')
        ]
    ]
    new_tables = [
        Table('spamfilter_statistics',
              key=('strategy', 'action', 'data', 'status'))[
            Column('strategy'),
            Column('action'),
            Column('data'),
            Column('status'),
            Column('delay', type='double precision'),
            Column('delay_max', type='double precision'),
            Column('delay_min', type='double precision'),
            Column('count', type='int'),
            Column('external', type='int'),
            Column('time', type='int')
        ],
        Table('spamfilter_report', key='id')[
            Column('id', auto_increment=True),
            Column('entry'),
            Column('headers'),
            Column('author'),
            Column('authenticated', type='int'),
            Column('comment'),
            Column('time', type='int')
        ]
    ]
    dbm = DatabaseManager(env)
    dbm.create_tables(new_tables)
    dbm.upgrade_tables(changed_tables)



# -*- coding: utf-8 -*-
#
# Copyright (C) 2005-2011 Edgewall Software
# Copyright (C) 2005 Matthew Good <trac@matt-good.net>
# Copyright (C) 2006 Christopher Lenz <cmlenz@gmx.de>
# Copyright (C) 2011 Dirk Stöcker <trac@dstoecker.de>
# All rights reserved.
#
# This software is licensed as described in the file COPYING, which
# you should have received as part of this distribution. The terms
# are also available at http://trac.edgewall.com/license.html.
#
# This software consists of voluntary contributions made by many
# individuals. For the exact contribution history, see the revision
# history and logs, available at http://projects.edgewall.com/trac/.
#
# Author: Dirk Stöcker <trac@dstoecker.de>,
#         Matthew Good <trac@matt-good.net>

import re

from trac.config import BoolOption, IntOption, Option
from trac.core import Component, implements
from trac.wiki.api import IWikiChangeListener
from trac.wiki.model import WikiPage

from tracspamfilter.api import IFilterStrategy, N_


class IPRegexFilterStrategy(Component):
    """Spam filter for submitter's IP based on regular expressions
    defined in BadIP page.
    """
    implements(IFilterStrategy, IWikiChangeListener)

    karma_points = IntOption('spam-filter', 'ipregex_karma', '20',
        """By how many points a match with a pattern on the BadIP page
        impacts the overall karma of a submission.""",
        doc_domain='tracspamfilter')

    badcontent_file = Option('spam-filter', 'ipbadcontent_file', '',
        """Local file to be loaded to get BadIP. Can be used in
        addition to BadIP wiki page.""", doc_domain='tracspamfilter')

    show_blacklisted = BoolOption('spam-filter', 'show_blacklisted_ip',
        'true',
        "Show the matched bad IP patterns in rejection message.",
        doc_domain='tracspamfilter')

    def __init__(self):
        self.patterns = []
        page = WikiPage(self.env, 'BadIP')
        if page.exists:
            self._load_patterns(page)
        if self.badcontent_file != '':
            file = open(self.badcontent_file, 'r')
            if file is None:
                self.log.warning("BadIP file cannot be opened")
            else:
                lines = file.read().splitlines()
                pat = [re.compile(p.strip()) for p in lines if p.strip()]
                self.log.debug("Loaded %s patterns from BadIP file", len(pat))
                self.patterns += pat

    # IFilterStrategy implementation

    def is_external(self):
        return False

    def test(self, req, author, content, ip):
        gotcha = []
        points = 0
        for pattern in self.patterns:
            match = pattern.search(ip)
            if match:
                gotcha.append("'%s'" % pattern.pattern)
                self.log.debug("Pattern %s found in submission",
                               pattern.pattern)
                points -= abs(self.karma_points)
        if points != 0:
            if self.show_blacklisted:
                matches = ", ".join(gotcha)
                return points, N_("IP catched by these blacklisted "
                                  "patterns: %s"), matches
            else:
                return (points, N_("IP catched by %s blacklisted patterns"),
                        str(len(gotcha)))

    def train(self, req, author, content, ip, spam=True):
        return 0

    # IWikiChangeListener implementation

    def wiki_page_changed(self, page, *args):
        if page.name == 'BadIP':
            self._load_patterns(page)
    wiki_page_added = wiki_page_changed
    wiki_page_version_deleted = wiki_page_changed

    def wiki_page_deleted(self, page):
        if page.name == 'BadIP':
            self.patterns = []

    # Internal methods

    def _load_patterns(self, page):
        if '{{{' in page.text and '}}}' in page.text:
            lines = page.text.split('{{{', 1)[1].split('}}}', 1)[0].splitlines()
            self.patterns = [re.compile(p.strip()) for p in lines if p.strip()]
            self.log.debug("Loaded %s patterns from BadIP",
                           len(self.patterns))
        else:
            self.log.warning("BadIP page does not contain any patterns")
            self.patterns = []



# -*- coding: utf-8 -*-
#
# Copyright (C) 2006 Edgewall Software
# All rights reserved.
#
# This software is licensed as described in the file COPYING, which
# you should have received as part of this distribution. The terms
# are also available at http://trac.edgewall.com/license.html.
#
# This software consists of voluntary contributions made by many
# individuals. For the exact contribution history, see the revision
# history and logs, available at http://projects.edgewall.com/trac/.

from acct_mgr.register import IAccountRegistrationInspector
from trac.core import Component, implements
from trac.util.html import tag
from trac.util.translation import cleandoc_

from tracspamfilter.api import _
from tracspamfilter.filters.trapfield import TrapFieldFilterStrategy
from tracspamfilter.filtersystem import FilterSystem
try:
    from tracspamfilter.filters.registration import RegistrationFilterStrategy
except ImportError:
    RegistrationFilterStrategy = None


class RegistrationFilterAdapter(Component):
    """Interface of spamfilter to account manager plugin to check new account
    registrations for spam."""

    _domain = 'tracspamfilter'
    _description = cleandoc_(
    """Interface of spamfilter to account manager plugin to check new account
    registrations for spam.

    It provides an additional 'Details' input field to get more information
    for calculating the probability of a spam registration attempt.
    Knowledge gained from inspecting registration attempts is shared with all
    other spam filter adapters for this system.
    """)
    implements(IAccountRegistrationInspector)

    # IAccountRegistrationInspector methods

    def render_registration_fields(self, req, data):
        insert = tag.div(tag.label(_("Details:"), tag.input(type="text",
                     name=TrapFieldFilterStrategy(self.env).name_register, size=60, class_="textwidget")))

        if RegistrationFilterStrategy:
            return RegistrationFilterStrategy(self.env).render_registration_fields(req, data, dict(optional=insert))
        else:
            return dict(optional=insert), data

    def validate_registration(self, req):
        # Add the author/reporter name
        if req.authname and req.authname != 'anonymous':
            author = req.authname
        else:
            author = req.args.get('username', req.authname)
        email = req.args.get('email')
        if email:
            author += " <%s>" % email

        changes = []
        sentinel = req.args.get('sentinel')
        if sentinel:
            changes += [(None, sentinel)]
        name = req.args.get('name')
        if name:
            changes += [(None, name)]
        comment = req.args.get(TrapFieldFilterStrategy(self.env).name_register)
        if comment:
            changes += [(None, comment)]

        FilterSystem(self.env).test(req, author, changes)
        return []



# -*- coding: utf-8 -*-
#
# Copyright (C) 2006 Edgewall Software
# Copyright (C) 2006 Alec Thomas <alec@swapoff.org>
# All rights reserved.
#
# This software is licensed as described in the file COPYING, which
# you should have received as part of this distribution. The terms
# are also available at http://trac.edgewall.com/license.html.
#
# This software consists of voluntary contributions made by many
# individuals. For the exact contribution history, see the revision
# history and logs, available at http://projects.edgewall.com/trac/.
#
# Author: Alec Thomas <alec@swapoff.org>

import StringIO
import os
import random

from PIL import Image
from PIL import ImageFont
from PIL import ImageDraw
from PIL import ImageFilter

from trac.config import ListOption, IntOption, Option
from trac.core import Component, TracError, implements
from trac.env import ISystemInfoProvider
from trac.util import get_pkginfo
from trac.util.html import html
from trac.web.api import IRequestHandler

from tracspamfilter.captcha.api import ICaptchaMethod


class ImageCaptcha(Component):
    """An image captcha courtesy of
    http://aspn.activestate.com/ASPN/Cookbook/Python/Recipe/440588
    """

    implements(ICaptchaMethod, IRequestHandler, ISystemInfoProvider)

    fonts = ListOption('spam-filter', 'captcha_image_fonts', 'vera.ttf',
        doc="""Set of fonts to choose from when generating image CAPTCHA.""",
        doc_domain='tracspamfilter')

    font_size = IntOption('spam-filter', 'captcha_image_font_size', 25,
        """Font size to use in image CAPTCHA.""",
        doc_domain='tracspamfilter')

    alphabet = Option('spam-filter', 'captcha_image_alphabet',
                      'abcdefghkmnopqrstuvwxyz',
        """Alphabet to choose image CAPTCHA challenge from.""",
        doc_domain='tracspamfilter')

    letters = IntOption('spam-filter', 'captcha_image_letters', 6,
        """Number of letters to use in image CAPTCHA challenge.""",
        doc_domain='tracspamfilter')

    # IRequestHandler methods
    def match_request(self, req):
        return req.path_info == '/captcha/image'

    def process_request(self, req):
        if 'captcha_expected' not in req.session:
            # TODO Probably need to render an error image here
            raise TracError("No CAPTCHA response in session")
        req.send_response(200)
        req.send_header('Content-Type', 'image/jpeg')

        image = StringIO.StringIO()
        from pkg_resources import resource_filename
        font = os.path.join(resource_filename('tracspamfilter', 'fonts'),
                            random.choice(self.fonts))
        self.gen_captcha(image, req.session['captcha_expected'], font,
                         self.font_size)
        img = image.getvalue()
        req.send_header('Content-Length', len(img))
        req.end_headers()
        req.write(img)

    # ICaptchaMethod methods

    def generate_captcha(self, req):
        challenge = ''.join(random.choice(self.alphabet)
                            for i in xrange(self.letters))
        return challenge, html.img(src=req.href('/captcha/image'),
                                   width='33%', alt='captcha')

    def verify_captcha(self, req):
        return False

    def is_usable(self, req):
        return True

    # ISystemInfoProvider methods

    def get_system_info(self):
        import PIL
        yield 'Pillow', get_pkginfo(PIL)['version']

    # Internal methods

    def gen_captcha(self, file, text, fnt, fnt_sz, fmt='JPEG'):
        # randomly select the foreground color
        fgcolor = random.randint(0, 0xffff00)
        # make the background color the opposite of fgcolor
        bgcolor = fgcolor ^ 0xffffff
        # create a font object
        font = ImageFont.truetype(fnt, fnt_sz)
        # determine dimensions of the text
        dim = font.getsize(text)
        # create a new image slightly larger that the text
        im = Image.new('RGB', (dim[0] + 5, dim[1] + 5), bgcolor)
        d = ImageDraw.Draw(im)
        x, y = im.size
        r = random.randint
        # draw 100 random colored boxes on the background
        for num in xrange(100):
            d.rectangle((r(0, x), r(0, y), r(0, x), r(0, y)),
                        fill=r(0, 0xffffff))
        # add the text to the image
        d.text((3, 3), text, font=font, fill=fgcolor)
        im = im.filter(ImageFilter.EDGE_ENHANCE_MORE)
        # save the image to a file
        im.save(file, format=fmt)



# -*- coding: utf-8 -*-
#
# Copyright (C) 2014 Edgewall Software
# All rights reserved.
#
# This software is licensed as described in the file COPYING, which
# you should have received as part of this distribution. The terms
# are also available at http://trac.edgewall.com/license.html.
#
# This software consists of voluntary contributions made by many
# individuals. For the exact contribution history, see the revision
# history and logs, available at http://projects.edgewall.com/trac/.

from trac.admin import IAdminPanelProvider
from trac.config import IntOption
from trac.core import Component, implements
from trac.web.api import HTTPNotFound
from trac.web.chrome import (
    add_link, add_script, add_script_data, add_stylesheet)

from tracspamfilter.api import _
from tracspamfilter.model import LogEntry


class ReportAdminPageProvider(Component):
    """Web administration panel for reviewing Spam reports"""

    implements(IAdminPanelProvider)

    MAX_PER_PAGE = 10000
    MIN_PER_PAGE = 5
    DEF_PER_PAGE =  IntOption('spam-filter', 'spam_report_entries', '100',
    "How many report entries are displayed by default (between 5 and 10000).",
    doc_domain='tracspamfilter')

    # IAdminPanelProvider methods

    def get_admin_panels(self, req):
        if 'SPAM_CHECKREPORTS' in req.perm:
            total = self.env.db_query("SELECT COUNT(*) FROM spamfilter_report")
            total = total[0][0]

            yield ('spamfilter', _("Spam Filtering"), 'report', _("Reports") \
                + (" (%s)"%total if total else ""))

    def render_admin_panel(self, req, cat, page, path_info):
        req.perm.require('SPAM_CHECKREPORTS')
        if req.method == 'POST' and 'delete' in req.args:
            entries = req.args.getlist('sel')
            if entries:
                self.env.db_transaction("""
                    DELETE FROM spamfilter_report WHERE id IN (%s)
                    """ % ",".join("'%s'" % each for each in entries))
            req.redirect(req.href.admin(cat, page,
                                        page=req.args.get('page'),
                                        num=req.args.get('num')))
        if path_info:
            data = self._render_entry(req, cat, page, path_info)
            page = 'entry'
            data['allowselect'] = False
            data['entries'] = LogEntry.selectrelated(self.env, data['path'], data['time'])
        else:
            data = self._render_panel(req, cat, page)
            page = ''

        add_stylesheet(req, 'spamfilter/admin.css')
        return 'admin_report%s.html' % page, data

    # Internal methods

    def _render_panel(self, req, cat, page):
        try:
            pagenum = int(req.args.get('page', 1)) - 1
        except ValueError:
            pagenum = 1

        try:
            pagesize = int(req.args.get('num', self.DEF_PER_PAGE))
        except ValueError:
            pagesize = self.DEF_PER_PAGE
        if pagesize < self.MIN_PER_PAGE:
            pagesize = self.MIN_PER_PAGE
        elif pagesize > self.MAX_PER_PAGE:
            pagesize = self.MAX_PER_PAGE

        total = self.env.db_query("SELECT COUNT(*) FROM spamfilter_report")
        total = total[0][0]

        if total < pagesize:
            pagenum = 0
        elif total <= pagenum * pagesize:
            pagenum = (total-1)/pagesize

        offset = pagenum * pagesize
        entries = []
        idx = 0;
        for e in self.env.db_query("""
            SELECT id,time,entry,author,authenticated,comment
            FROM spamfilter_report
            ORDER BY time DESC LIMIT %s OFFSET %s""", (pagesize, offset)):
            # don't display additional appended values
            p = e[2].split("#")
            entries.append(('odd' if idx %2 else 'even',)+e[0:2]+(p[0],)+e[3:])
            idx += 1;

        if pagenum > 0:
            add_link(req, 'prev', req.href.admin(cat, page, page=pagenum,
                                                 num=pagesize),
                     _('Previous Page'))
        if offset + pagesize < total:
            add_link(req, 'next', req.href.admin(cat, page, page=pagenum+2,
                                                 num=pagesize),
                     _('Next Page'))

        if entries:
            add_script_data(req, {'toggleform': "spamreportform"})
        return {
            'entries': entries,
            'offset': offset + 1,
            'page': pagenum + 1,
            'num': pagesize,
            'total': total
        }

    def _render_entry(self, req, cat, page, entry_id):
        with self.env.db_query as db:
            entry = db("""
                SELECT time,entry,author,authenticated,headers,comment
                FROM spamfilter_report
                WHERE id = %s""", (entry_id,))
            if not entry:
                raise HTTPNotFound(_('Report entry not found'))
            entry = entry[0]

            for previous, in db("""
                SELECT id
                FROM spamfilter_report
                WHERE id<%s ORDER BY id DESC LIMIT 1""", (entry_id,)):
                add_link(req, 'prev', req.href.admin(cat, page, previous),
                         _('Report Entry %d') % previous)
            add_link(req, 'up', req.href.admin(cat, page), _('Report Entry List'))
            for next, in db("""
                SELECT id
                FROM spamfilter_report
                WHERE id>%s ORDER BY id DESC LIMIT 1""", (entry_id,)):
                add_link(req, 'next', req.href.admin(cat, page, next),
                         _('Report Entry %d') % next)

            # don't display additional appended values
            path = entry[1].split("#")
            return {'time': entry[0],
                    'monitor': 'SPAM_MONITOR' in req.perm,
                    'id': entry_id,
                    'path': path[0],
                    'author': entry[2],
                    'authenticated': entry[3],
                    'headers': entry[4],
                    'comment': entry[5]}



# -*- coding: utf-8 -*-
#
# Copyright (C) 2015 Edgewall Software
# Copyright (C) 2015 Dirk Stöcker <trac@dstoecker.de>
# All rights reserved.
#
# This software is licensed as described in the file COPYING, which
# you should have received as part of this distribution. The terms
# are also available at http://trac.edgewall.com/license.html.
#
# This software consists of voluntary contributions made by many
# individuals. For the exact contribution history, see the revision
# history and logs, available at http://projects.edgewall.com/trac/.
#
# Author: Dirk Stöcker <trac@dstoecker.de>

import urllib2

from trac.admin import IAdminPanelProvider
from trac.config import BoolOption, IntOption
from trac.core import Component, implements
from trac.web.api import HTTPNotFound
from trac.web.chrome import (
    ITemplateProvider, add_link, add_script, add_script_data, add_stylesheet)

from tracspamfilter.api import _, gettext, ngettext
from tracspamfilter.filters.akismet import AkismetFilterStrategy
from tracspamfilter.filters.botscout import BotScoutFilterStrategy
from tracspamfilter.filters.fspamlist import FSpamListFilterStrategy
from tracspamfilter.filters.stopforumspam import StopForumSpamFilterStrategy
from tracspamfilter.filtersystem import FilterSystem
from tracspamfilter.model import LogEntry, Statistics
try:
    from tracspamfilter.filters.bayes import BayesianFilterStrategy
except ImportError:  # SpamBayes not installed
    BayesianFilterStrategy = None
try:
    from tracspamfilter.filters.httpbl import HttpBLFilterStrategy
    from tracspamfilter.filters.ip_blacklist import IPBlacklistFilterStrategy
    from tracspamfilter.filters.url_blacklist import URLBlacklistFilterStrategy
except ImportError:  # DNS python not installed
    HttpBLFilterStrategy = None
    IPBlacklistFilterStrategy = None
    URLBlacklistFilterStrategy = None


class SpamFilterAdminPageProvider(Component):
    """Web administration panel for configuring and monitoring the spam
    filtering system.
    """

    implements(ITemplateProvider)
    implements(IAdminPanelProvider)

    MAX_PER_PAGE = 10000
    MIN_PER_PAGE = 5
    DEF_PER_PAGE = IntOption('spam-filter', 'spam_monitor_entries', '100',
            "How many monitor entries are displayed by default "
            "(between 5 and 10000).", doc_domain='tracspamfilter')

    train_only = BoolOption('spam-filter', 'show_train_only', False,
            "Show the buttons for training without deleting entry.",
            doc_domain='tracspamfilter')

    # IAdminPanelProvider methods

    def get_admin_panels(self, req):
        if 'SPAM_CONFIG' in req.perm:
            yield ('spamfilter', _("Spam Filtering"),
                   'config', _("Configuration"))
        if 'SPAM_MONITOR' in req.perm:
            yield ('spamfilter', _("Spam Filtering"),
                   'monitor', _("Monitoring"))

    def render_admin_panel(self, req, cat, page, path_info):
        if page == 'config':
            if req.method == 'POST':
                if self._process_config_panel(req):
                    req.redirect(req.href.admin(cat, page))
            data = self._render_config_panel(req, cat, page)
        else:
            if req.method == 'POST':
                if self._process_monitoring_panel(req):
                    req.redirect(req.href.admin(cat, page,
                                                page=req.args.getint('page'),
                                                num=req.args.getint('num')))
            if path_info:
                data = self._render_monitoring_entry(req, cat, page, path_info)
                page = 'entry'
            else:
                data = self._render_monitoring_panel(req, cat, page)
                data['allowselect'] = True
                data['monitor'] = True
                add_script_data(req, {
                    'bayestext': _("SpamBayes determined spam probability "
                                   "of %s%%"),
                    'sel100text': _("Select 100.00%% entries") % (),
                    'sel90text': _("Select &gt;90.00%% entries") % (),
                    'sel10text': _("Select &lt;10.00%% entries") % (),
                    'sel0text': _("Select 0.00%% entries") % (),
                    'selspamtext': _("Select Spam entries"),
                    'selhamtext': _('Select Ham entries'),
                    'nolog_obvious': FilterSystem(self.env).nolog_obvious
                })
                add_script(req, 'spamfilter/adminmonitor.js')
                add_script_data(req, {'toggleform': 'spammonitorform'})

        add_stylesheet(req, 'spamfilter/admin.css')
        data['accmgr'] = 'ACCTMGR_USER_ADMIN' in req.perm
        return 'admin_spam%s.html' % page, data

    # ITemplateProvider methods

    def get_htdocs_dirs(self):
        from pkg_resources import resource_filename
        return [('spamfilter', resource_filename(__name__, 'htdocs'))]

    def get_templates_dirs(self):
        from pkg_resources import resource_filename
        return [resource_filename(__name__, 'templates')]

    # Internal methods

    def _render_config_panel(self, req, cat, page):
        req.perm.require('SPAM_CONFIG')
        filter_system = FilterSystem(self.env)

        strategies = []
        for strategy in filter_system.strategies:
            for variable in dir(strategy):
                if variable.endswith('karma_points'):
                    strategies.append({
                        'name': strategy.__class__.__name__,
                        'karma_points': getattr(strategy, variable),
                        'variable': variable,
                        'karma_help': gettext(getattr(strategy.__class__,
                                                      variable).__doc__)
                    })

        add_script(req, 'spamfilter/adminconfig.js')
        return {
            'strategies': sorted(strategies, key=lambda x: x['name']),
            'min_karma': filter_system.min_karma,
            'authenticated_karma': filter_system.authenticated_karma,
            'attachment_karma': filter_system.attachment_karma,
            'register_karma': filter_system.register_karma,
            'trust_authenticated': filter_system.trust_authenticated,
            'logging_enabled': filter_system.logging_enabled,
            'nolog_obvious': filter_system.nolog_obvious,
            'purge_age': filter_system.purge_age,
            'spam_monitor_entries_min': self.MIN_PER_PAGE,
            'spam_monitor_entries_max': self.MAX_PER_PAGE,
            'spam_monitor_entries': self.DEF_PER_PAGE
        }

    def _process_config_panel(self, req):
        req.perm.require('SPAM_CONFIG')

        spam_config = self.config['spam-filter']

        min_karma = req.args.as_int('min_karma')
        if min_karma is not None:
            spam_config.set('min_karma', min_karma)

        attachment_karma = req.args.as_int('attachment_karma')
        if attachment_karma is not None:
            spam_config.set('attachment_karma', attachment_karma)

        register_karma = req.args.as_int('register_karma')
        if register_karma is not None:
            spam_config.set('register_karma', register_karma)

        authenticated_karma = req.args.as_int('authenticated_karma')
        if authenticated_karma is not None:
            spam_config.set('authenticated_karma', authenticated_karma)

        for strategy in FilterSystem(self.env).strategies:
            for variable in dir(strategy):
                if variable.endswith('karma_points'):
                    key = strategy.__class__.__name__ + '_' + variable
                    points = req.args.get(key)
                    if points is not None:
                        option = getattr(strategy.__class__, variable)
                        self.config.set(option.section, option.name, points)

        logging_enabled = 'logging_enabled' in req.args
        spam_config.set('logging_enabled', logging_enabled)

        nolog_obvious = 'nolog_obvious' in req.args
        spam_config.set('nolog_obvious', nolog_obvious)

        trust_authenticated = 'trust_authenticated' in req.args
        spam_config.set('trust_authenticated', trust_authenticated)

        if logging_enabled:
            purge_age = req.args.as_int('purge_age')
            if purge_age is not None:
                spam_config.set('purge_age', purge_age)

        spam_monitor_entries = req.args.as_int('spam_monitor_entries',
                                               min=self.MIN_PER_PAGE,
                                               max=self.MAX_PER_PAGE)
        if spam_monitor_entries is not None:
            spam_config.set('spam_monitor_entries', spam_monitor_entries)

        self.config.save()
        return True

    def _render_monitoring_panel(self, req, cat, page):
        req.perm.require('SPAM_MONITOR')

        pagenum = req.args.as_int('page', 1) - 1

        pagesize = req.args.as_int('num', self.DEF_PER_PAGE,
                                   min=self.MIN_PER_PAGE,
                                   max=self.MAX_PER_PAGE)

        total = LogEntry.count(self.env)

        if total < pagesize:
            pagenum = 0
        elif total <= pagenum * pagesize:
            pagenum = (total - 1) / pagesize

        offset = pagenum * pagesize
        entries = list(LogEntry.select(self.env, limit=pagesize,
                                       offset=offset))
        if pagenum > 0:
            add_link(req, 'prev',
                     req.href.admin(cat, page, page=pagenum, num=pagesize),
                     _("Previous Page"))
        if offset + pagesize < total:
            add_link(req, 'next',
                     req.href.admin(cat, page, page=pagenum + 2, num=pagesize),
                     _("Next Page"))

        return {
            'enabled': FilterSystem(self.env).logging_enabled,
            'entries': entries,
            'offset': offset + 1,
            'page': pagenum + 1,
            'num': pagesize,
            'total': total,
            'train_only': self.train_only
        }

    def _render_monitoring_entry(self, req, cat, page, entry_id):
        req.perm.require('SPAM_MONITOR')

        entry = LogEntry.fetch(self.env, entry_id)
        if not entry:
            raise HTTPNotFound(_("Log entry not found"))

        previous = entry.get_previous()
        if previous:
            add_link(req, 'prev', req.href.admin(cat, page, previous.id),
                     _("Log Entry %(id)s", id=previous.id))
        add_link(req, 'up', req.href.admin(cat, page), _("Log Entry List"))
        next = entry.get_next()
        if next:
            add_link(req, 'next', req.href.admin(cat, page, next.id),
                     _("Log Entry %(id)s", id=next.id))

        return {'entry': entry, 'train_only': self.train_only}

    def _process_monitoring_panel(self, req):
        req.perm.require('SPAM_TRAIN')

        filtersys = FilterSystem(self.env)

        spam = 'markspam' in req.args or 'markspamdel' in req.args
        train = spam or 'markham' in req.args or 'markhamdel' in req.args
        delete = 'delete' in req.args or 'markspamdel' in req.args or \
                 'markhamdel' in req.args or 'deletenostats' in req.args
        deletestats = 'delete' in req.args

        if train or delete:
            entries = req.args.getlist('sel')
            if entries:
                if train:
                    filtersys.train(req, entries, spam=spam, delete=delete)
                elif delete:
                    filtersys.delete(req, entries, deletestats)

        if 'deleteobvious' in req.args:
            filtersys.deleteobvious(req)

        return True


class ExternalAdminPageProvider(Component):
    """Web administration panel for configuring the External spam filters."""

    implements(IAdminPanelProvider)

    # IAdminPanelProvider methods

    def get_admin_panels(self, req):
        if 'SPAM_CONFIG' in req.perm:
            yield ('spamfilter', _("Spam Filtering"),
                   'external', _("External Services"))

    def render_admin_panel(self, req, cat, page, path_info):
        req.perm.require('SPAM_CONFIG')

        data = {}
        spam_config = self.config['spam-filter']

        akismet = AkismetFilterStrategy(self.env)
        stopforumspam = StopForumSpamFilterStrategy(self.env)
        botscout = BotScoutFilterStrategy(self.env)
        fspamlist = FSpamListFilterStrategy(self.env)

        ip_blacklist_default = ip6_blacklist_default = \
            url_blacklist_default = None
        if HttpBLFilterStrategy:
            ip_blacklist = IPBlacklistFilterStrategy(self.env)
            ip_blacklist_default = ip_blacklist.servers_default
            ip6_blacklist_default = ip_blacklist.servers6_default
            url_blacklist = URLBlacklistFilterStrategy(self.env)
            url_blacklist_default = url_blacklist.servers_default

        if req.method == 'POST':
            if 'cancel' in req.args:
                req.redirect(req.href.admin(cat, page))

            akismet_api_url = req.args.get('akismet_api_url')
            akismet_api_key = req.args.get('akismet_api_key')
            stopforumspam_api_key = req.args.get('stopforumspam_api_key')
            botscout_api_key = req.args.get('botscout_api_key')
            fspamlist_api_key = req.args.get('fspamlist_api_key')
            httpbl_api_key = req.args.get('httpbl_api_key')
            ip_blacklist_servers = req.args.get('ip_blacklist_servers')
            ip6_blacklist_servers = req.args.get('ip6_blacklist_servers')
            url_blacklist_servers = req.args.get('url_blacklist_servers')
            use_external = 'use_external' in req.args
            train_external = 'train_external' in req.args
            skip_external = req.args.get('skip_external')
            stop_external = req.args.get('stop_external')
            skip_externalham = req.args.get('skip_externalham')
            stop_externalham = req.args.get('stop_externalham')
            try:
                verified_key = akismet.verify_key(req, akismet_api_url,
                                                  akismet_api_key)
                if akismet_api_key and not verified_key:
                    data['akismeterror'] = 'The API key is invalid'
                    data['error'] = 1
            except urllib2.URLError, e:
                data['alismeterror'] = e.reason[1]
                data['error'] = 1

            if not data.get('error', 0):
                spam_config.set('akismet_api_url', akismet_api_url)
                spam_config.set('akismet_api_key', akismet_api_key)
                spam_config.set('stopforumspam_api_key', stopforumspam_api_key)
                spam_config.set('botscout_api_key', botscout_api_key)
                spam_config.set('fspamlist_api_key', fspamlist_api_key)
                spam_config.set('httpbl_api_key', httpbl_api_key)
                if HttpBLFilterStrategy:
                    if ip_blacklist_servers != ip_blacklist_default:
                        spam_config.set('ip_blacklist_servers',
                                        ip_blacklist_servers)
                    else:
                        spam_config.remove('ip_blacklist_servers')
                    if ip6_blacklist_servers != ip6_blacklist_default:
                        spam_config.set('ip6_blacklist_servers',
                                        ip6_blacklist_servers)
                    else:
                        spam_config.remove('ip6_blacklist_servers')
                    if url_blacklist_servers != url_blacklist_default:
                        spam_config.set('url_blacklist_servers',
                                        url_blacklist_servers)
                    else:
                        spam_config.remove('url_blacklist_servers')

                spam_config.set('use_external', use_external)
                spam_config.set('train_external', train_external)
                spam_config.set('skip_external', skip_external)
                spam_config.set('stop_external', stop_external)
                spam_config.set('skip_externalham', skip_externalham)
                spam_config.set('stop_externalham', stop_externalham)
                self.config.save()
                req.redirect(req.href.admin(cat, page))

        else:
            filter_system = FilterSystem(self.env)
            use_external = filter_system.use_external
            train_external = filter_system.train_external
            skip_external = filter_system.skip_external
            stop_external = filter_system.stop_external
            skip_externalham = filter_system.skip_externalham
            stop_externalham = filter_system.stop_externalham
            akismet_api_url = akismet.api_url
            akismet_api_key = akismet.api_key
            stopforumspam_api_key = stopforumspam.api_key
            botscout_api_key = botscout.api_key
            fspamlist_api_key = fspamlist.api_key
            httpbl_api_key = spam_config.get('httpbl_api_key')
            ip_blacklist_servers = spam_config.get('ip_blacklist_servers')
            ip6_blacklist_servers = spam_config.get('ip6_blacklist_servers')
            url_blacklist_servers = spam_config.get('url_blacklist_servers')

        if HttpBLFilterStrategy:
            data['blacklists'] = 1
            data['ip_blacklist_default'] = ip_blacklist_default
            data['ip6_blacklist_default'] = ip6_blacklist_default
            data['url_blacklist_default'] = url_blacklist_default

        data.update({
            'akismet_api_key': akismet_api_key,
            'akismet_api_url': akismet_api_url,
            'httpbl_api_key': httpbl_api_key,
            'stopforumspam_api_key': stopforumspam_api_key,
            'botscout_api_key': botscout_api_key,
            'fspamlist_api_key': fspamlist_api_key,
            'use_external': use_external,
            'train_external': train_external,
            'skip_external': skip_external,
            'stop_external': stop_external,
            'skip_externalham': skip_externalham,
            'stop_externalham': stop_externalham,
            'ip_blacklist_servers': ip_blacklist_servers,
            'ip6_blacklist_servers': ip6_blacklist_servers,
            'url_blacklist_servers': url_blacklist_servers
        })

        add_script(req, 'spamfilter/adminexternal.js')
        add_stylesheet(req, 'spamfilter/admin.css')
        return 'admin_external.html', data


class BayesAdminPageProvider(Component):
    """Web administration panel for configuring the Bayes spam filter."""

    if BayesianFilterStrategy:
        implements(IAdminPanelProvider)

    # IAdminPanelProvider methods

    def get_admin_panels(self, req):
        if 'SPAM_CONFIG' in req.perm:
            yield 'spamfilter', _("Spam Filtering"), 'bayes', _("Bayes")

    def render_admin_panel(self, req, cat, page, path_info):
        req.perm.require('SPAM_CONFIG')

        bayes = BayesianFilterStrategy(self.env)
        hammie = bayes._get_hammie()
        data = {}

        if req.method == 'POST':
            if 'train' in req.args:
                bayes.train(None, None, req.args['bayes_content'], '127.0.0.1',
                            spam='spam' in req.args['train'].lower())
                req.redirect(req.href.admin(cat, page))

            elif 'test' in req.args:
                bayes_content = req.args['bayes_content']
                data['content'] = bayes_content
                try:
                    data['score'] = hammie.score(bayes_content.encode('utf-8'))
                except Exception, e:
                    self.log.warn('Bayes test failed: %s', e, exc_info=True)
                    data['error'] = unicode(e)

            else:
                if 'reset' in req.args:
                    self.log.info('Resetting SpamBayes training database')
                    self.env.db_transaction("DELETE FROM spamfilter_bayes")
                elif 'reduce' in req.args:
                    self.log.info('Reducing SpamBayes training database')
                    bayes.reduce()

                min_training = req.args.as_int('min_training')
                if min_training is not None and \
                        min_training != bayes.min_training:
                    self.config.set('spam-filter', 'bayes_min_training',
                                    min_training)
                    self.config.save()

                min_dbcount = req.args.as_int('min_dbcount')
                if min_dbcount is not None and \
                        min_dbcount != bayes.min_dbcount:
                    self.config.set('spam-filter', 'bayes_min_dbcount',
                                    min_dbcount)
                    self.config.save()

                req.redirect(req.href.admin(cat, page))
        ratio = ''
        nspam = hammie.bayes.nspam
        nham = hammie.bayes.nham
        if nham and nspam:
            if nspam > nham:
                ratio = _("(ratio %.1f : 1)") % (float(nspam) / float(nham))
            else:
                ratio = _("(ratio 1 : %.1f)") % (float(nham) / float(nspam))

        dblines, dblines_spamonly, dblines_hamonly, dblines_reduce = \
            bayes.dblines()
        dblines_mixed = dblines - dblines_hamonly - dblines_spamonly
        data.update({
            'min_training': bayes.min_training,
            'min_dbcount': bayes.min_dbcount,
            'dblines': dblines,
            'dblinesreducenum': dblines_reduce,
            'dblinesspamonly':
                ngettext("%(num)d spam", "%(num)d spam", dblines_spamonly),
            'dblineshamonly':
                ngettext("%(num)d ham", "%(num)d ham", dblines_hamonly),
            'dblinesreduce':
                ngettext("%(num)d line", "%(num)d lines", dblines_reduce),
            'dblinesmixed':
                ngettext("%(num)d mixed", "%(num)d mixed", dblines_mixed),
            'nspam': nspam,
            'nham': nham,
            'ratio': ratio
        })

        add_script_data(req, {'hasdata': True if nham + nspam > 0 else False})
        add_script(req, 'spamfilter/adminbayes.js')
        add_stylesheet(req, 'spamfilter/admin.css')
        return 'admin_bayes.html', data


class StatisticsAdminPageProvider(Component):
    """Web administration panel for spam filter statistics."""

    implements(IAdminPanelProvider)

    # IAdminPanelProvider methods

    def get_admin_panels(self, req):
        if 'SPAM_CONFIG' in req.perm:
            yield ('spamfilter', _("Spam Filtering"),
                   'statistics', _("Statistics"))

    def render_admin_panel(self, req, cat, page, path_info):
        req.perm.require('SPAM_CONFIG')

        stats = Statistics(self.env)

        if req.method == 'POST':
            if 'clean' in req.args:
                stats.clean(req.args['strategy'])
            elif 'cleanall' in req.args:
                stats.cleanall()
            req.redirect(req.href.admin(cat, page))

        strategies, overall = stats.getstats()

        data = {'strategies': strategies, 'overall': overall}

        add_stylesheet(req, 'spamfilter/admin.css')
        return 'admin_statistics.html', data



# -*- coding: utf-8 -*-
#
# Copyright (C) 2006 Edgewall Software
# All rights reserved.
#
# This software is licensed as described in the file COPYING, which
# you should have received as part of this distribution. The terms
# are also available at http://trac.edgewall.com/license.html.
#
# This software consists of voluntary contributions made by many
# individuals. For the exact contribution history, see the revision
# history and logs, available at http://projects.edgewall.com/trac/.

from datetime import datetime, timedelta

from trac.config import IntOption
from trac.core import Component, implements
from trac.util.datefmt import datetime_now

from tracspamfilter.api import IFilterStrategy, N_
from tracspamfilter.model import LogEntry


class IPThrottleFilterStrategy(Component):
    """Spam filter strategy that throttles multiple subsequent submissions from
    the same IP address.
    """
    implements(IFilterStrategy)

    karma_points = IntOption('spam-filter', 'ip_throttle_karma', '3',
        """By how many points exceeding the configured maximum number of
        posts per hour impacts the overall score.""",
        doc_domain='tracspamfilter')

    max_posts = IntOption('spam-filter', 'max_posts_by_ip', '10',
        """The maximum allowed number of submissions per hour from a single
        IP address. If this limit is exceeded, subsequent submissions get
        negative karma.""", doc_domain='tracspamfilter')

    # IFilterStrategy implementation

    def is_external(self):
        return False

    def test(self, req, author, content, ip):
        threshold = datetime_now() - timedelta(hours=1)
        num_posts = 0

        for entry in LogEntry.select(self.env, ipnr=ip):
            if datetime.fromtimestamp(entry.time) < threshold:
                break
            num_posts += 1

        if num_posts > self.max_posts:
            return -abs(self.karma_points) * num_posts / self.max_posts, \
                   N_("Maximum number of posts per hour for this IP exceeded")

    def train(self, req, author, content, ip, spam=True):
        return 0



# -*- coding: utf-8 -*-
#
# Copyright (C) 2006 Edgewall Software
# All rights reserved.
#
# This software is licensed as described in the file COPYING, which
# you should have received as part of this distribution. The terms
# are also available at http://trac.edgewall.com/license.html.
#
# This software consists of voluntary contributions made by many
# individuals. For the exact contribution history, see the revision
# history and logs, available at http://projects.edgewall.com/trac/.

import binascii
import re
import xml.etree.ElementTree as ElementTree
from datetime import datetime, timedelta
from time import mktime, time

from trac.db import Column, Index, Table
from trac.util import as_int
from trac.util.text import to_unicode
from trac.util.datefmt import datetime_now

from tracspamfilter.api import gettext, _, get_strategy_name


class LogEntry(object):

    table = Table('spamfilter_log', key='id')[
        Column('id', auto_increment=True),
        Column('time', type='int'),
        Column('path'),
        Column('author'),
        Column('authenticated', type='int'),
        Column('ipnr'),
        Column('headers'),
        Column('content'),
        Column('rejected', type='int'),
        Column('karma', type='int'),
        Column('reasons'),
        Column('request')
    ]

    def __init__(self, env, time, path, author, authenticated, ipnr, headers,
                 content, rejected, karma, reasons, request):
        self.id = None
        self.env = env
        self.time = time
        self.path = path
        self.author = to_unicode(author)
        self.authenticated = bool(authenticated)
        self.ipnr = ipnr
        self.headers = to_unicode(headers) or ''
        self.content = content
        self.rejected = bool(rejected)
        self.karma = karma
        self.reasons = reasons
        self.request = request

    def __repr__(self):
        date = datetime.fromtimestamp(self.time).isoformat()
        return '<%s %s from %s by "%s">' % (self.__class__.__name__, self.id,
                                            date, self.author)

    exists = property(fget=lambda self: self.id is not None,
                      doc='Whether this log entry exists in the database')

    @classmethod
    def _encode_content(cls, content):
        """Take a `basestring` content and return a plain text encoding."""
        return to_unicode(content).encode('utf-8').encode('base64')

    @classmethod
    def _decode_content(cls, content):
        """Revert the encoding done by `_encode_content` and return an unicode
        string"""
        try:
            return to_unicode(content.decode('base64'))
        except (UnicodeEncodeError, binascii.Error):
            # cope with legacy content (stored before base64 encoding)
            return to_unicode(content)

    def get_next(self):
        """Return the next log entry in reverse chronological order (i.e. the
        next older entry.)"""
        for row in self.env.db_query("""
                SELECT id,time,path,author,authenticated,ipnr,headers,
                       content,rejected,karma,reasons,request
                FROM spamfilter_log
                WHERE id<%s ORDER BY id DESC LIMIT 1
                """, (self.id,)):
            return self.__class__._from_db(self.env, row)

    def get_previous(self):
        """Return the previous log entry in reverse chronological order
        (i.e. the next younger entry.)"""
        for row in self.env.db_query("""
                SELECT id,time,path,author,authenticated,ipnr,headers,
                       content,rejected,karma,reasons,request
                FROM spamfilter_log
                WHERE id>%s ORDER BY id LIMIT 1
                """, (self.id,)):
            return self.__class__._from_db(self.env, row)

    def insert(self):
        """Insert a new log entry into the database."""
        assert not self.exists, 'Cannot insert existing log entry'

        content = self._encode_content(self.content)
        with self.env.db_transaction as db:
            cursor = db.cursor()
            cursor.execute("""
                INSERT INTO spamfilter_log
                 (time,path,author,authenticated,ipnr,headers,content,
                  rejected,karma,reasons,request)
                VALUES (%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s)
                """, (int(self.time), self.path, self.author,
                      int(bool(self.authenticated)), self.ipnr, self.headers,
                      content, int(bool(self.rejected)), int(self.karma),
                      self._reasons_to_xml(self.reasons),
                      self._request_to_xml(self.request)))
            self.id = db.get_last_id(cursor, 'spamfilter_log')

    def update(self, **kwargs):
        """Update the log entry in the database."""
        assert self.exists, 'Cannot update a non-existing log entry'

        for name, value in kwargs.items():
            if hasattr(self, name):
                setattr(self, name, value)

        content = self._encode_content(self.content)
        self.env.db_transaction("""
            UPDATE spamfilter_log
            SET time=%s,path=%s,author=%s,authenticated=%s,ipnr=%s,headers=%s,
                content=%s,rejected=%s,karma=%s,reasons=%s,request=%s
            WHERE id=%s
            """, (int(self.time), self.path, self.author,
                  int(bool(self.authenticated)), self.ipnr, self.headers,
                  content, int(bool(self.rejected)), int(self.karma),
                  self._reasons_to_xml(self.reasons),
                  self._request_to_xml(self.request), self.id))

    @classmethod
    def delete(cls, env, id):
        """Delete the log entry with the specified ID from the database."""
        if isinstance(id, list):
            env.db_transaction.executemany("""
                DELETE FROM spamfilter_log WHERE id=%s
                """, [(i,) for i in id])
        else:
            env.db_transaction("""
                DELETE FROM spamfilter_log WHERE id=%s
                """, (id,))

    @classmethod
    def getobvious(cls, env):
        """Delete the obvious log entries from the database."""
        ids = []
        for id, reasons in env.db_query("""
                SELECT id,reasons
                FROM spamfilter_log
                WHERE authenticated=0 AND rejected=1
                """):
            for r in cls._xml_to_reasons(reasons):
                if r[0] == 'Bayesian' and float(r[3]) > 90:
                    ids.append(id)
        return ids

    @classmethod
    def fetch(cls, env, id):
        """Retrieve an existing log entry from the database by ID."""
        for row in env.db_query("""
                SELECT id,time,path,author,authenticated,ipnr,headers,
                       content,rejected,karma,reasons,request
                FROM spamfilter_log
                WHERE id=%s
                """, (int(id),)):
            return cls._from_db(env, row)

    @classmethod
    def count(cls, env):
        """Return the number of log entries in the database."""
        for count, in env.db_query("SELECT COUNT(*) FROM spamfilter_log "):
            return count

    @classmethod
    def purge(cls, env, days):
        """Delete log entries older than the specified number of days."""
        threshold = datetime_now() - timedelta(days=days)
        env.db_transaction("""
            DELETE FROM spamfilter_log WHERE time < %s
            """, (mktime(threshold.timetuple()),))

    @classmethod
    def select(cls, env, ipnr=None, limit=None, offset=0):
        """Retrieve existing log entries from the database that match the
        specified criteria.
        """
        params = []
        where_clauses = []
        if ipnr:
            where_clauses.append("ipnr=%s")
            params.append(ipnr)

        if where_clauses:
            where = "WHERE %s" % " AND ".join(where_clauses)
        else:
            where = ""

        return cls._doselect(env, where, limit, offset, params)

    @classmethod
    def selectrelated(cls, env, page, limit=None, offset=0):
        """Retrieve page related log entries"""
        # strip additional appended values
        p = page.split('#')
        params = [p[0]]
        where_clauses = ['path = %s']

        if page.startswith('/ticket/'):
            if len(p) == 2:  # if we have time, use it
                modtime = p[1]
            else:  # otherwise hope ticket is not yet deleted
                modtime = None
                for t, in env.db_query("""
                        SELECT time/1000000 from ticket WHERE id=%s
                        """, (page[8:],)):
                    modtime = t
            if modtime:
                where_clauses.append("(path = '/newticket' AND "
                                     "time BETWEEN %s AND %s)")
                params.append(modtime-30)
                params.append(modtime+30)

        where = "WHERE %s" % " OR ".join(where_clauses)
        return cls._doselect(env, where, limit, offset, params)

    @classmethod
    def _doselect(cls, env, where, limit, offset, params):
        extra_clauses = []

        if limit:
            extra_clauses.append("LIMIT %s")
            params.append(limit)
            if offset:
                extra_clauses.append("OFFSET %s")
                params.append(offset)
        if extra_clauses:
            extra = " ".join(extra_clauses)
        else:
            extra = ""

        for row in env.db_query("""
                SELECT id,time,path,author,authenticated,ipnr,headers,
                       content,rejected,karma,reasons,request
                FROM spamfilter_log
                %s
                ORDER BY time DESC %s
                """ % (where, extra), params):
            yield cls._from_db(env, row)

    @classmethod
    def _from_db(cls, env, row):
        """Create a new LogEntry from a row from the `spamfilter_log` table.
        """
        fields = list(row[1:])
        fields[6] = cls._decode_content(fields[6])
        fields[9] = cls._xml_to_reasons(fields[9])
        fields[10] = cls._xml_to_request(fields[10])
        obj = cls(env, *fields)
        obj.id = row[0]
        return obj

    def _reasons_to_xml(self, reasons):
        root = ElementTree.Element("entries")
        for r in reasons:
            e = ElementTree.SubElement(root, 'e', name=r[0], points=r[1],
                                       text=r[2])
            for a in r[3:]:
                ElementTree.SubElement(e, 'v', value=a)
        return ElementTree.tostring(root)

    def _request_to_xml(self, request):
        if request is None:
            return None
        try:
            root = ElementTree.Element('request', target=request[0])
            for key, value in request[1].iteritems():
                ElementTree.SubElement(root, 'arg', key=key).text = value
            return ElementTree.tostring(root)
        except Exception:
            return None

    @classmethod
    def _xml_to_reasons(self, xml):
        reasons = []
        if xml:
            try:
                root = ElementTree.fromstring(xml)
            except ElementTree.ParseError:
                # which is probably not XML but a plain msg
                # e.g.  u'BlogSpam (-5): BlogSpam says content is spam (
                # Blacklisted name, coco)'
                m = re.match(r'(\w+) \((-?\d+)\): [^(]+ \((.*)\)', xml)
                if m:
                    reasons.append(m.groups())
            else:
                for r in list(root):
                    reasons.append([r.attrib.get('name'),
                                    r.attrib.get('points'),
                                    r.attrib.get('text')] +
                                   [a.attrib.get('value') for a in list(r)])
        return reasons

    @classmethod
    def _xml_to_request(self, xml):
        if xml is None:
            return None
        try:
            root = ElementTree.fromstring(xml)
            request = [root.attrib.get('target'), {}]
            for r in list(root):
                request[1][r.attrib.get('key')] = r.text or ''
            return request
        except Exception:
            return None

    def getreasons(self):
        reasons = []
        for r in self.reasons:
            if len(r) == 3:
                reasons.append("%s (%s): %s" % (r[0], r[1], gettext(r[2])))
            else:
                reasons.append("%s (%s): %s"
                               % (r[0], r[1], gettext(r[2]) % tuple(r[3:])))
        return reasons

    def findreasonvalue(self, name):
        for r in self.reasons:
            if r[0] == name:
                return as_int(r[1], 0)
        return 0


class Bayes(object):

    table = Table('spamfilter_bayes', key='word')[
        Column('word'),
        Column('nspam', type='int'),
        Column('nham', type='int'),
        Index(['word'])
    ]


class Statistics(object):

    table = Table('spamfilter_statistics',
                  key=('strategy', 'action', 'data', 'status'))[
        Column('strategy'),
        Column('action'),
        Column('data'),
        Column('status'),
        Column('delay', type='double precision'),
        Column('delay_max', type='double precision'),
        Column('delay_min', type='double precision'),
        Column('count', type='int'),
        Column('external', type='int'),
        Column('time', type='int')
    ]

    def __init__(self, env):
        self.env = env

    def insert_or_update(self, strategy, action, data, status, delay,
                         external):
        with self.env.db_transaction as db:
            row = db("""
                    SELECT delay,delay_max,delay_min,count
                    FROM spamfilter_statistics
                    WHERE action=%s AND data=%s AND status=%s AND strategy=%s
                    """, (action, data, status, strategy))
            if row:
                count = float(row[0][3])
                delay_val = float(row[0][0]) * (count / (count + 1.0)) + \
                            delay / (count + 1.0)
                delay_max = float(row[0][1])
                if delay > delay_max:
                    delay_max = delay
                delay_min = float(row[0][2])
                if delay < delay_min:
                    delay_min = delay
                count = int(row[0][3]) + 1
                db("""
                    UPDATE spamfilter_statistics
                    SET delay=%s,delay_max=%s,delay_min=%s,count=%s
                    WHERE action=%s AND data=%s AND status=%s AND strategy=%s
                    """, (delay_val, delay_max, delay_min, count, action,
                          data, status, strategy))
            else:
                db("""
                    INSERT INTO spamfilter_statistics
                    VALUES (%s,%s,%s,%s,%s,%s,%s,1,%s,%s)
                    """, (strategy, action, data, status, delay, delay,
                          delay, external, int(time())))
        self.env.log.debug("SPAMLOG: %s %s %s %s %.3f %s", action, data,
                           status, strategy, delay,
                           "external" if external else "local")

    def clean(self, strategy):
        self.env.db_transaction("""
                DELETE FROM spamfilter_statistics WHERE strategy=%s
                """, (strategy,))

    def cleanall(self):
        self.env.db_transaction("DELETE FROM spamfilter_statistics")

    def getstats(self):
        strategies = {}
        overall = {'test': 0, 'time': 0, 'testspam': 0, 'testham': 0,
                   'testint': 0, 'testext': 0, 'testinttime': 0,
                   'testexttime': 0, 'spamerror': 0, 'hamerror': 0}
        for (strategy, action, data, status, delay, delay_max, delay_min,
             count, external, time) in \
                self.env.db_query("SELECT * FROM spamfilter_statistics"):
            if strategy:
                props = strategies.get(strategy, {})
                props['type'] = "external" if external else "internal"
                props['i18type'] = \
                    _("external") if external else _("internal")
                if action == 'test':
                    if data == 'empty':
                        props['testempty'] = count
                    elif data == 'ham' or data == 'spam':
                        props['test%s%s' % (data, status)] = count
                    lasttotal = props.get('testtotal', 0)
                    props['testtotal'] = count + lasttotal
                    if lasttotal:
                        l = float(lasttotal)
                        c = float(count)
                        s = l + c
                        props['testtime'] = props['testtime'] * (l / s) + \
                                          float(delay) * (c / s)
                    else:
                        props['testtime'] = float(delay)
                elif action.startswith('train') or action == 'delete':
                    if status:
                        key = 'train%s%s' % (data, status)
                        props[key] = count + props.get(key, 0)
                        key = 'train%stotal' % data
                        props[key] = count + props.get(key, 0)
                        props['traintotal'] = count + \
                                              props.get('traintotal', 0)
                    if action == 'trainerror' or action == 'traincond':
                        props[action] = count + props.get(action, 0)
                strategies[strategy] = props
            else:
                if not overall['time'] or time < overall['time']:
                    overall['time'] = time
                if (action == 'trainint' or action == 'trainext') and \
                        status == 'error':
                    if data == 'spam':
                        overall['spamerror'] += count
                    else:
                        overall['hamerror'] += count
                if action == 'testint' or action == 'testext':
                    overall[action] += count
                    overall['%stime' % action] = delay
                    overall['test'] += count
                    if data == 'spam':
                        overall['testspam'] += count
                    else:
                        overall['testham'] += count
        return strategies, overall

    def sortbyperformance(self, entries):
        strategies = {}
        for strategy, action, data, status, delay, count in \
                self.env.db_query("""
                    SELECT strategy,action,data,status,delay,count
                    FROM spamfilter_statistics
                    WHERE external=1 AND strategy != ''
                    """):
            props = strategies.get(strategy, {
                'count': 0, 'delay': 0, 'ham': False,
                'err': 0, 'empty': 0, 'testcount': 0
            })
            if action == 'delete' or action.startswith('train'):
                if status == 'error':
                    props['err'] += count
                if status:
                    props['count'] += count
            elif action == 'test':
                if delay > props['delay']:
                    props['delay'] = delay
                if data == 'empty':
                    props['empty'] += count
                elif data == 'ham':
                    props['ham'] = True
                props['testcount'] += count
            strategies[strategy] = props
        for strategy, props in strategies.iteritems():
            # 10% if ham is reported
            c = 10 if props['ham'] else 0
            # 40% if fast
            s = 40 - int(props['delay'] * 20)
            c += s if s > 0 else 0
            # 25% if low error rate
            if props['count']:
                e = 25 - int(1000.0 * props['err'] / props['count'])
            else:
                e = 15
            c += e if e > 0 else 0
            # 25% if high answer rate
            if props['testcount']:
                a = 25 - int(25.0 * props['empty'] / props['testcount'])
            else:
                a = 15
            c += a if a > 0 else 0
            strategies[strategy] = c

        def mysort(val):
            return strategies.get(get_strategy_name(val), 0)

        entries = sorted(entries, key=mysort, reverse=True)
        return entries


class SpamReport(object):
    table = Table('spamfilter_report', key='id')[
        Column('id', auto_increment=True),
        Column('entry'),
        Column('headers'),
        Column('author'),
        Column('authenticated', type='int'),
        Column('comment'),
        Column('time', type='int')
    ]

schema = [Bayes.table, LogEntry.table, Statistics.table, SpamReport.table]
schema_version = 4
schema_version_name = 'spamfilter_version'



# -*- coding: utf-8 -*-
#
# Copyright (C) 2005-2006 Edgewall Software
# Copyright (C) 2005 Matthew Good <trac@matt-good.net>
# Copyright (C) 2006 Christopher Lenz <cmlenz@gmx.de>
# All rights reserved.
#
# This software is licensed as described in the file COPYING, which
# you should have received as part of this distribution. The terms
# are also available at http://trac.edgewall.com/license.html.
#
# This software consists of voluntary contributions made by many
# individuals. For the exact contribution history, see the revision
# history and logs, available at http://projects.edgewall.com/trac/.
#
# Author: Matthew Good <trac@matt-good.net>

from trac.config import IntOption
from trac.core import Component, implements
from trac.env import ISystemInfoProvider
from trac.util import get_pkginfo
from tracspamfilter.api import IFilterStrategy, N_

from spambayes.hammie import Hammie
from spambayes.storage import SQLClassifier


class BayesianFilterStrategy(Component):
    """Bayesian filtering strategy based on SpamBayes."""

    implements(IFilterStrategy, ISystemInfoProvider)

    karma_points = IntOption('spam-filter', 'bayes_karma', '15',
        """By what factor Bayesian spam probability score affects the overall
        karma of a submission.""", doc_domain='tracspamfilter')

    min_training = IntOption('spam-filter', 'bayes_min_training', '25',
        """The minimum number of submissions in the training database required
        for the filter to start impacting the karma of submissions.""",
        doc_domain='tracspamfilter')

    min_dbcount = IntOption('spam-filter', 'bayes_min_dbcount', '5',
        """Entries with a count less than this value get removed from the
        database when calling the reduce function.""",
        doc_domain='tracspamfilter')

    # IFilterStrategy implementation

    def is_external(self):
        return False

    def test(self, req, author, content, ip):
        hammie = self._get_hammie()
        nspam = hammie.bayes.nspam
        nham = hammie.bayes.nham
        if author is not None:
            testcontent = author + '\n' + content
        else:
            testcontent = content

        if min(nspam, nham) < self.min_training:
            self.log.info("Bayes filter strategy requires more training. "
                          "It currently has only %d words marked as ham, and "
                          "%d marked as spam, but requires at least %d for "
                          "each.", nham, nspam, self.min_training)
            return

        if nham - nspam > min(nham, nspam) * 2:
            self.log.warn("The difference between the number of ham versus "
                          "spam submissions in the training database is "
                          "large, results may be bad.")

        score = hammie.score(testcontent.encode('utf-8'))
        self.log.debug("SpamBayes reported spam probability of %s", score)
        points = -int(round(self.karma_points * (score * 2 - 1)))
        if points != 0:
            return (points,
                    N_("SpamBayes determined spam probability of %s%%"),
                    ("%3.2f" % (score * 100)))

    def train(self, req, author, content, ip, spam=True):
        if author is not None:
            testcontent = author + '\n' + content
        else:
            testcontent = content
        self.log.info("Training SpamBayes, marking content as %s",
                      spam and "spam" or "ham")

        hammie = self._get_hammie()
        hammie.train(testcontent.encode('utf-8', 'ignore'), spam)
        hammie.store()
        return 1

    # ISystemInfoProvider methods

    def get_system_info(self):
        import spambayes
        yield 'SpamBayes', get_pkginfo(spambayes)['version']

    # Internal methods

    def _get_hammie(self):
        try:  # 1.0
            return Hammie(TracDbClassifier(self.env, self.log))
        except TypeError:  # 1.1
            return Hammie(TracDbClassifier(self.env, self.log), 'c')

    def _get_numbers(self):
        hammie = self._get_hammie()
        return hammie.nspam, hammie.nham

    # used by admin panel

    def reduce(self):
        self.env.db_transaction("""
            DELETE FROM spamfilter_bayes
            WHERE nspam+nham < %s AND NOT word = 'saved state'
            """, (self.min_dbcount,))

    def dblines(self):
        total = self.env.db_query("""
            SELECT COUNT(*) FROM spamfilter_bayes
            WHERE NOT word = 'saved state'
            """)[0][0]
        spam = self.env.db_query("""
            SELECT COUNT(*) FROM spamfilter_bayes
            WHERE nham = 0 AND NOT word = 'saved state'
            """)[0][0]
        ham = self.env.db_query("""
            SELECT COUNT(*) FROM spamfilter_bayes
            WHERE nspam = 0 AND NOT word = 'saved state'
            """)[0][0]
        reduce = self.env.db_query("""
            SELECT COUNT(*) FROM spamfilter_bayes
            WHERE nspam+nham < %s AND NOT word = 'saved state'
            """, (self.min_dbcount,))[0][0]
        return total, spam, ham, reduce


class TracDbClassifier(SQLClassifier):
    # FIXME: This thing is incredibly slow

    def __init__(self, env_db, log):
        self.env_db = env_db
        self.log = log
        self.nham = None
        self.nspam = None
        SQLClassifier.__init__(self, 'Trac')

    def load(self):
        if self._has_key(self.statekey):
            row = self._get_row(self.statekey)
            self.nspam = row['nspam']
            self.nham = row['nham']
        else:  # new database
            self.nspam = self.nham = 0

    def _sanitize(self, text):
        if isinstance(text, unicode):
            return text
        # Remove invalid byte sequences from utf-8 encoded text
        return text.decode('utf-8', 'ignore')

    def _get_row(self, word):
        word = self._sanitize(word)
        for row in self.env_db.db_query("""
                SELECT nspam,nham FROM spamfilter_bayes WHERE word=%s
                """, (word,)):
            break
        else:
            return {}
        # prevent assertion - happens when there are failures in training and
        # the count is not updated due to an exception
        if word != self.statekey:
            if row[0] > self.nspam:
                self.log.warn("Reset SPAM count from %d to %d due to keyword "
                              "'%s'.", self.nspam, row[0], word)
                self.nspam = row[0]
                self.store()
            if row[1] > self.nham:
                self.log.warn("Reset HAM count from %d to %d due to keyword "
                              "'%s'.", self.nham, row[1], word)
                self.nham = row[1]
                self.store()
        return {'nspam': row[0], 'nham': row[1]}

    def _set_row(self, word, nspam, nham):
        word = self._sanitize(word)
        with self.env_db.db_transaction as db:
            if self._has_key(word):
                db("UPDATE spamfilter_bayes SET nspam=%s,nham=%s "
                   "WHERE word=%s", (nspam, nham, word))
            else:
                db("INSERT INTO spamfilter_bayes (word,nspam,nham) "
                   "VALUES (%s,%s,%s)", (word, nspam, nham))

    def _delete_row(self, word):
        word = self._sanitize(word)
        self.env_db.db_transaction("""
            DELETE FROM spamfilter_bayes WHERE word=%s
            """, (word,))

    def _has_key(self, key):
        key = self._sanitize(key)
        for count, in self.env_db.db_query("""
                SELECT COUNT(*) FROM spamfilter_bayes WHERE word=%s
                """, (key,)):
            return bool(count)

    def _wordinfoget(self, word):
        row = self._get_row(word)
        if row:
            item = self.WordInfoClass()
            item.__setstate__((row['nspam'], row['nham']))
            return item

    def _wordinfokeys(self):
        words = []
        for word, in self.env_db.db_query("""
                SELECT word FROM spamfilter_bayes
                """):
            words.append(word)
        return words



# -*- coding: utf-8 -*-
#
# Copyright (C) 2005-2012 Edgewall Software
# Copyright (C) 2005-2006 Matthew Good <trac@matt-good.net>
# Copyright (C) 2006 Christopher Lenz <cmlenz@gmx.de>
# All rights reserved.
#
# This software is licensed as described in the file COPYING, which
# you should have received as part of this distribution. The terms
# are also available at http://trac.edgewall.com/license.html.
#
# This software consists of voluntary contributions made by many
# individuals. For the exact contribution history, see the revision
# history and logs, available at http://projects.edgewall.com/trac/.
#
# Author: Matthew Good <trac@matt-good.net>
#         Christopher Lenz <cmlenz@gmx.de>

import time
from StringIO import StringIO
from difflib import SequenceMatcher
from pkg_resources import resource_filename

from trac.config import BoolOption, ConfigSection, ExtensionOption, IntOption
from trac.core import Component, ExtensionPoint, implements
from trac.db import DatabaseManager
from trac.env import IEnvironmentSetupParticipant, ISystemInfoProvider
from trac.perm import IPermissionRequestor
from trac.util import get_pkginfo, as_int
from trac.util.html import tag
from trac.util.text import shorten_line, to_unicode
from trac.web.api import Request

from tracspamfilter.api import (
    IFilterStrategy, IRejectHandler, RejectContent,
    _, N_, add_domain, get_strategy_name, gettext, tag_)
from tracspamfilter.filters.trapfield import TrapFieldFilterStrategy
from tracspamfilter.model import LogEntry, Statistics, schema, \
                                 schema_version, schema_version_name


class FilterSystem(Component):
    """The central component for spam filtering. Must be enabled always to allow
    filtering of spam.
    """

    strategies = ExtensionPoint(IFilterStrategy)

    implements(IEnvironmentSetupParticipant, IPermissionRequestor,
               IRejectHandler, ISystemInfoProvider)

    min_karma = IntOption('spam-filter', 'min_karma', '0',
        """The minimum score required for a submission to be allowed.""",
        doc_domain='tracspamfilter')

    authenticated_karma = IntOption('spam-filter', 'authenticated_karma', '20',
        """The karma given to authenticated users, in case
        `trust_authenticated` is false.""", doc_domain='tracspamfilter')

    logging_enabled = BoolOption('spam-filter', 'logging_enabled', 'true',
        """Whether all content submissions and spam filtering activity should
        be logged to the database.""", doc_domain='tracspamfilter')

    nolog_obvious = BoolOption('spam-filter', 'nolog_obvious', 'false',
        """Don't log obvious spam (Bayes >90%%, marked as spam and not
        authenticated) to the database.""", doc_domain='tracspamfilter')

    purge_age = IntOption('spam-filter', 'purge_age', '7',
        """The number of days after which log entries should be purged.""",
        doc_domain='tracspamfilter')

    use_external = BoolOption('spam-filter', 'use_external', 'true',
        """Allow usage of external services.""", doc_domain='tracspamfilter')

    skip_external = IntOption('spam-filter', 'skip_external', '20',
        """Skip external calls when this negative karma is already reached
        by internal tests.""", doc_domain='tracspamfilter')

    skip_externalham = IntOption('spam-filter', 'skip_externalham', '30',
        """Skip external calls when this positive karma is already reached
        by internal tests.""", doc_domain='tracspamfilter')

    stop_external = IntOption('spam-filter', 'stop_external', '50',
        """Stop external calls when this negative karma is reached.""",
        doc_domain='tracspamfilter')

    stop_externalham = IntOption('spam-filter', 'stop_externalham', '50',
        """Stop external calls when this positive karma is reached.""",
        doc_domain='tracspamfilter')

    train_external = BoolOption('spam-filter', 'train_external', 'true',
        """Allow training of external services.""", doc_domain='tracspamfilter')

    trust_authenticated = BoolOption('spam-filter', 'trust_authenticated',
                                     'false',
        """Whether content submissions by authenticated users should be trusted
        without checking for potential spam or other abuse.""",
        doc_domain='tracspamfilter')

    attachment_karma = IntOption('spam-filter', 'attachment_karma', '0',
        """The karma given to attachments.""", doc_domain='tracspamfilter')

    register_karma = IntOption('spam-filter', 'register_karma', '0',
        """The karma given to registrations.""", doc_domain='tracspamfilter')

    reject_handler = ExtensionOption('spam-filter', 'reject_handler',
                                     IRejectHandler, 'FilterSystem',
        """The handler used to reject content.""", doc_domain='tracspamfilter')

    isforwarded = BoolOption('spam-filter', 'is_forwarded', 'false',
        """Interpret X-Forwarded-For header for IP checks.""",
        doc_domain='tracspamfilter')

    spam_section = ConfigSection('spam-filter',
        """This section is used to handle all configurations used by
        spam filter plugin.""", doc_domain='tracspamfilter')

    def __init__(self):
        """Set up translation domain"""
        try:
            locale_dir = resource_filename(__name__, 'locale')
        except KeyError:
            pass
        else:
            add_domain(self.env.path, locale_dir)

    # IRejectHandler methods

    def reject_content(self, req, message):
        raise RejectContent(message)

    # Public methods

    def test(self, req, author, changes):
        """Test a submission against the registered filter strategies.

        @param req: the request object
        @param author: the name of the logged in user, or 'anonymous' if the
            user is not logged in
        @param changes: a list of `(old_content, new_content)` tuples for every
            modified "field", where `old_content` may contain the previous
            version of that field (if available), and `new_content` contains
            the newly submitted content
        """

        start = time.time()
        ip = req.remote_addr
        if self.isforwarded:
            x_forwarded = req.get_header('X-Forwarded-For')
            if x_forwarded and x_forwarded != '':
                ip = x_forwarded.split(',',1)[0]

        if author.find("@") < 1:
            trap = TrapFieldFilterStrategy(self.env).get_trap(req)
            if trap:
                if trap.find("@") > 0:
                    author += " <%s>" % trap
                else:
                    self.log.debug("Append trap field to changes")
                    changes.append((None, trap))

        score = 0
        if self.trust_authenticated:
            # Authenticated users are trusted
            if req.authname and req.authname != 'anonymous':
                return

        reasons = []
        outreasons = []
        results = []
        if req.authname and req.authname != 'anonymous':
            reasons.append(("AuthenticatedUserScore", str(self.authenticated_karma),
                            N_("User is authenticated")))
            score += self.authenticated_karma

        if self.attachment_karma != 0 and \
                req.args.get('attachment') is not None:
            reasons.append(("AttachmentScore", str(self.attachment_karma),
                            N_("Attachment weighting")))
            score += self.attachment_karma

        if self.register_karma != 0 and (req.path_info == "/register" or req.args.get(TrapFieldFilterStrategy(self.env).name_register)):
            reasons.append(("RegisterScore", str(self.register_karma),
                            N_("Registration weighting")))
            score += self.register_karma

        if not author:
            author = 'anonymous'
        self.log.debug("Spam testing for %s", req.path_info)
        content = self._combine_changes(changes)
        abbrev = shorten_line(content)
        self.log.debug('Testing content %r submitted by "%s"', abbrev, author)
        externals = []

        obvious = False
        for strategy in self.strategies:
            tim = time.time()
            try:
                if not strategy.is_external():
                    retval = strategy.test(req, author, content, ip)
                    tim = time.time()-tim
                    if tim > 3:
                        self.log.warn('Test %s took %d seconds to complete.',
                                      strategy, tim)
                    if retval:
                        points = retval[0]
                        if len(retval) > 2:
                            reason = retval[1] % retval[2:]
                        else:
                            reason = retval[1]
                        if points < 0:
                            if len(retval) > 2:
                                outreasons.append(gettext(retval[1]) % retval[2:])
                            else:
                                outreasons.append(gettext(retval[1]))

                        self.log.debug('Filter strategy %r gave submission %d '
                                       'karma points (reason: %r)', strategy,
                                       points, reason)
                        score += points
                        if reason:
                            name = get_strategy_name(strategy)
                            if name == "Bayesian" and float(retval[2]) > 90:
                                obvious = True
                            reasons.append((name, str(points)) + retval[1:])
                            results.append((strategy, points, tim))
                    else:
                        self._record_action('test','empty', '', strategy, tim)
                elif self.use_external:
                    externals.append(strategy)
            except Exception, e:
                self._record_action('test', 'error', '', strategy, time.time()-tim)
                self.log.exception('Filter strategy %s raised exception: %s', strategy, e)

        extint = "testint"
        if score > -self.skip_external and score < self.skip_externalham:
            externals = Statistics(self.env).sortbyperformance(externals)
            for strategy in externals:
                tim = time.time()
                try:
                    if score > -self.stop_external and score < self.stop_externalham:
                        extint = "testext"
                        retval = strategy.test(req, author, content, ip)
                        tim = time.time()-tim
                        if tim > 3:
                            self.log.warn('Test %s took %d seconds to '
                                          'complete.', strategy, tim)
                        if retval:
                            points = retval[0]
                            if len(retval) > 2:
                                reason = retval[1] % retval[2:]
                            else:
                                reason = retval[1]
                            if points < 0:
                                if len(retval) > 2:
                                    outreasons.append(gettext(retval[1]) % retval[2:])
                                else:
                                    outreasons.append(gettext(retval[1]))

                            self.log.debug('Filter strategy %r gave submission %d '
                                           'karma points (reason: %r)', strategy,
                                           points, reason)
                            score += points
                            if reason:
                                name = get_strategy_name(strategy)
                                reasons.append((name,str(points)) + retval[1:])
                                results.append((strategy, points, tim))
                        else:
                            self._record_action('test','empty', '', strategy, tim)
                except Exception, e:
                    self._record_action('test','error', '', strategy, time.time()-tim)
                    self.log.exception('Filter strategy %s raised exception: %s', strategy, e)

        reasons = sorted(reasons, key=lambda r: r[0])

        if score < self.min_karma:
            type = "spam"
            self._record_action(extint, 'spam', '', '', time.time()-start)
        else:
            type = "ham"
            self._record_action(extint, 'ham', '', '', time.time()-start)
        for strategy,points,tim in results:
            if (points < 0 and score < self.min_karma) or (points > 0 and score >= self.min_karma):
                status = "ok"
            else:
                status = "error"
            self._record_action("test", type, status, strategy, tim)

        if self.logging_enabled:
            headers = '\n'.join(['%s: %s' % (k[5:].replace('_', '-').title(), v)
                                 for k, v in req.environ.items()
                                 if k.startswith('HTTP_')])
            authenticated = req.authname and req.authname != 'anonymous'
            rejected = score < self.min_karma
            if not self.nolog_obvious or not obvious or authenticated or not rejected:
                LogEntry(self.env, time.time(), req.path_info, author,
                         authenticated, ip, headers, content, rejected,
                         score, reasons, [req.path_info, req.args]).insert()
            else:
                self.log.debug('Skip %s %s %d', author, req.remote_addr, score)
                type = "spam" if rejected else "ham"
                status = "delete"
                for strategy in self.strategies:
                    count = 0
                    name = get_strategy_name(strategy)
                    for r in reasons:
                        if r[0] == name:
                            count = as_int(r[1], 0)
                            break
                    if count:
                        spamstatus = count < 0
                        self._record_action(status, type, ("ok" if spamstatus == rejected else "error"), strategy, 0)
                    else:
                        self._record_action(status, type, '', strategy, 0)
                self._record_action(status, type, '', '', 0)

            LogEntry.purge(self.env, self.purge_age)

        if score < self.min_karma:
            self.log.debug('Rejecting submission %r by "%s" (%r) because it '
                           'earned only %d karma points (%d are required).',
                           abbrev, author, req.remote_addr, score, self.min_karma)
            rejects = []
            outreasons.sort()
            for r in outreasons:
                rejects.append(tag.li(r))
            msg = tag.div(tag.ul(rejects), class_='message')

            self.reject_handler.reject_content(req, msg)

    def train(self, req, ids, spam=True, delete=False):
        environ = {}
        for name, value in req.environ.items():
            if not name.startswith('HTTP_'):
                environ[name] = value

        if not isinstance(ids, list):
            ids = [ids]
        for log_id in ids:
            start = time.time()
            entry = LogEntry.fetch(self.env, log_id)
            if entry:
                extint = "trainint"
                self.log.debug('Marking as %s: %r submitted by "%s"',
                               spam and 'spam' or 'ham',
                               shorten_line(entry.content),
                               entry.author)
                fakeenv = environ.copy()
                for header in entry.headers.splitlines():
                    name, value = header.split(':', 1)
                    if name == 'Cookie': # breaks SimpleCookie somehow
                        continue
                    cgi_name = 'HTTP_%s' % name.strip().replace('-', '_').upper()
                    fakeenv[cgi_name] = value.strip()
                fakeenv['REQUEST_METHOD'] = 'POST'
                fakeenv['PATH_INFO'] = entry.path
                fakeenv['wsgi.input'] = StringIO('')
                fakeenv['REMOTE_ADDR'] = entry.ipnr
                if entry.authenticated:
                    fakeenv['REMOTE_USER'] = entry.author

                type = "spam" if spam else "ham"
                for strategy in self.strategies:
                    status = "trainskip"
                    if (self.use_external and self.train_external) or not strategy.is_external():
                        tim = time.time()
                        extint = "trainext"
                        res = strategy.train(Request(fakeenv, None),
                                   entry.author or 'anonymous',
                                   entry.content, entry.ipnr, spam=spam)
                        tim = time.time()-tim
                        if tim > 3:
                            self.log.warn('Training %s took %d seconds to '
                                          'complete.', strategy, tim)
                        if res == -1:
                            status = "trainerror"
                        elif res == -2:
                            status = "traincond"
                        elif res > 0:
                            status = "train"
                    count = entry.findreasonvalue(get_strategy_name(strategy))
                    if count:
                        spamstatus = count < 0
                        self._record_action(status, type, ("ok" if spamstatus == spam else "error"), strategy, tim)
                    else:
                        self._record_action(status, type, '', strategy, tim)

                self._record_action(extint, type, ("ok" if entry.rejected == spam else "error"), '', time.time()-start)
                entry.update(rejected=spam)
                if delete:
                    self.delete(req, log_id, True)

    def delete(self, req, ids, stats):
        if not isinstance(ids, list):
            ids = [ids]
        if stats:
            for log_id in ids:
                entry = LogEntry.fetch(self.env, log_id)
                if entry:
                    type = "spam" if entry.rejected else "ham"
                    status = "delete"
                    for strategy in self.strategies:
                        count = entry.findreasonvalue(get_strategy_name(strategy))
                        if count:
                            spamstatus = count < 0
                            self._record_action(status, type, ("ok" if spamstatus == entry.rejected else "error"), strategy, 0)
                        else:
                            self._record_action(status, type, '', strategy, 0)
                    self._record_action(status, type, '', '', 0)
        LogEntry.delete(self.env, ids)

    def deleteobvious(self, req):
        ids = LogEntry.getobvious(self.env)
        if ids:
            self.delete(req, ids, True)

    # IEnvironmentSetupParticipant methods

    def environment_created(self):
        self.upgrade_environment()

    def environment_needs_upgrade(self):
        dbm = DatabaseManager(self.env)
        return dbm.needs_upgrade(schema_version, schema_version_name)

    def upgrade_environment(self):
        dbm = DatabaseManager(self.env)
        current_version = dbm.get_database_version(schema_version_name)
        if not current_version:
            dbm.create_tables(schema)
            dbm.set_database_version(schema_version, schema_version_name)
        else:
            dbm.upgrade(schema_version, schema_version_name,
                        'tracspamfilter.upgrades')

    # IPermissionRequestor methods

    def get_permission_actions(self):
        perms = ['SPAM_CONFIG', 'SPAM_MONITOR', 'SPAM_TRAIN', 'SPAM_USER',
                'SPAM_REPORT', 'SPAM_CHECKREPORTS']
        return perms + [('SPAM_ADMIN', perms)]

    # ISystemInfoProvider methods

    def get_system_info(self):
        # Move implementation to httpbl, ip_blacklist and url_blacklist
        # when support for Trac < 1.2 is dropped (#12294).
        try:
            import dns
        except ImportError:
            pass
        else:
            yield 'dnspython', get_pkginfo(dns)['version']

    # Internal methods

    def _combine_changes(self, changes, sep='\n\n'):
        fields = []
        for old_content, new_content in changes:
            new_content = to_unicode(new_content)
            if old_content:
                old_content = to_unicode(old_content)
                new_content = self._get_added_lines(old_content, new_content)
            fields.append(new_content)
        return sep.join(fields)

    def _get_added_lines(self, old_content, new_content):
        buf = []
        old_lines = old_content.splitlines()
        new_lines = new_content.splitlines()
        matcher = SequenceMatcher(None, old_lines, new_lines)
        for group in matcher.get_grouped_opcodes(0):
            for tag, i1, i2, j1, j2 in group:
                if tag in ('insert', 'replace'):
                    buf.append('\n'.join(new_lines[j1:j2]))

        return '\n'.join(buf)

    def _record_action(self, action, data, status, strategy, delay):
        stats = Statistics(self.env)
        if strategy:
            name = get_strategy_name(strategy)
            stats.insert_or_update(name, action, data, status, delay, 1 if strategy.is_external() else 0)
        else:
            stats.insert_or_update('', action, data, status, delay, None)



# -*- coding: utf-8 -*-
#
# Copyright (C) 2006 Edgewall Software
# All rights reserved.
#
# This software is licensed as described in the file COPYING, which
# you should have received as part of this distribution. The terms
# are also available at http://trac.edgewall.com/license.html.
#
# This software consists of voluntary contributions made by many
# individuals. For the exact contribution history, see the revision
# history and logs, available at http://projects.edgewall.com/trac/.

import unittest

from trac.test import EnvironmentStub, MockRequest

from tracspamfilter.filters.extlinks import ExternalLinksFilterStrategy


class ExternalLinksFilterStrategyTestCase(unittest.TestCase):

    def setUp(self):
        self.env = EnvironmentStub(enable=[ExternalLinksFilterStrategy])
        self.strategy = ExternalLinksFilterStrategy(self.env)

    def test_no_links(self):
        req = MockRequest(self.env)
        retval = self.strategy.test(req, 'John Doe', 'Foo bar', '127.0.0.1')
        self.assertEqual(None, retval)

    def test_few_ext_links(self):
        req = MockRequest(self.env)
        retval = self.strategy.test(req, 'John Doe', """
        <a href="http://spammers-site.com/fakehandbags">fakehandbags</a>
        <a href="http://spammers-site.com/fakewatches">fakewatches</a>
        """, '127.0.0.1')
        self.assertEqual(None, retval)

    def test_many_ext_links(self):
        req = MockRequest(self.env)
        retval = self.strategy.test(req, 'John Doe', """
        <a href="http://spammers-site.com/fakehandbags">fakehandbags</a>
        <a href="http://spammers-site.com/fakewatches">fakewatches</a>
        <a href="http://spammers-site.com/fakehandbags">fakehandbags</a>
        <a href="http://spammers-site.com/fakewatches">fakewatches</a>
        <a href="http://spammers-site.com/fakehandbags">fakehandbags</a>
        <a href="http://spammers-site.com/fakewatches">fakewatches</a>
        """, '127.0.0.1')
        self.assertEqual(
            (-3, 'Maximum number of external links per post exceeded'),
            retval
        )

    def test_many_ext_links_same_site(self):
        req = MockRequest(self.env)
        retval = self.strategy.test(req, 'John Doe', """
        <a href="http://example.org/page1">foo</a>
        <a href="http://example.org/page2">bar</a>
        <a href="http://example.org/page1">foo</a>
        <a href="http://example.org/page2">bar</a>
        <a href="http://example.org/page1">foo</a>
        <a href="http://example.org/page2">bar</a>
        """, '127.0.0.1')
        self.assertEqual(None, retval)

    def test_many_ext_links_raw(self):
        req = MockRequest(self.env)
        retval = self.strategy.test(req, 'John Doe', """
        http://spammers-site.com/fakehandbags
        http://spammers-site.com/fakewatches
        http://spammers-site.com/fakehandbags
        http://spammers-site.com/fakewatches
        http://spammers-site.com/fakehandbags
        http://spammers-site.com/fakewatches
        """, '127.0.0.1')
        self.assertEqual(
            (-3, 'Maximum number of external links per post exceeded'),
            retval
        )

    def test_many_ext_links_bbcode(self):
        req = MockRequest(self.env)
        retval = self.strategy.test(req, 'John Doe', """
        [url=http://spammers-site.com/fakehandbags]fakehandbags[/url]
        [url=http://spammers-site.com/fakewatches]fakewatches[/url]
        [url=http://spammers-site.com/fakehandbags]fakehandbags[/url]
        [url=http://spammers-site.com/fakewatches]fakewatches[/url]
        [url=http://spammers-site.com/fakehandbags]fakehandbags[/url]
        [url=http://spammers-site.com/fakewatches]fakewatches[/url]
        """, '127.0.0.1')
        self.assertEqual(
            (-3, 'Maximum number of external links per post exceeded'),
            retval
        )


def test_suite():
    suite = unittest.TestSuite()
    suite.addTest(unittest.makeSuite(ExternalLinksFilterStrategyTestCase))
    return suite

if __name__ == '__main__':
    unittest.main(defaultTest='test_suite')



# -*- coding: utf-8 -*-
#
# Copyright (C) 2006 Edgewall Software
# All rights reserved.
#
# This software is licensed as described in the file COPYING, which
# you should have received as part of this distribution. The terms
# are also available at http://trac.edgewall.com/license.html.
#
# This software consists of voluntary contributions made by many
# individuals. For the exact contribution history, see the revision
# history and logs, available at http://projects.edgewall.com/trac/.

import unittest

from tracspamfilter.filters.tests import akismet, bayes, extlinks, regex, \
                                         session


def test_suite():
    suite = unittest.TestSuite()
    suite.addTest(akismet.test_suite())
    suite.addTest(bayes.test_suite())
    suite.addTest(extlinks.test_suite())
    suite.addTest(regex.test_suite())
    suite.addTest(session.test_suite())
    return suite

if __name__ == '__main__':
    unittest.main(defaultTest='test_suite')



# -*- coding: utf-8 -*-
#
# Copyright (C) 2011 Dirk Stöcker <trac@dstoecker.de>
# All rights reserved.
#
# This software is licensed as described in the file COPYING, which
# you should have received as part of this distribution. The terms
# are also available at http://trac.edgewall.com/license.html.
#
# This software consists of voluntary contributions made by many
# individuals. For the exact contribution history, see the revision
# history and logs, available at http://projects.edgewall.com/trac/.

import urllib2
from email.utils import parseaddr
from urllib import urlencode
from xml.etree import ElementTree

from trac.config import IntOption, Option
from trac.core import Component, implements

from tracspamfilter.api import IFilterStrategy, N_, user_agent


class StopForumSpamFilterStrategy(Component):
    """Spam filter using the StopForumSpam service (http://stopforumspam.com/).
    """
    implements(IFilterStrategy)

    karma_points = IntOption('spam-filter', 'stopforumspam_karma', '4',
        """By how many points a StopForumSpam reject impacts the overall
        karma of a submission.""", doc_domain='tracspamfilter')

    api_key = Option('spam-filter', 'stopforumspam_api_key', '',
        "API key used to report SPAM.", doc_domain='tracspamfilter')

    # IFilterStrategy implementation

    def is_external(self):
        return True

    def test(self, req, author, content, ip):
        if not self._check_preconditions(False):
            return
        try:
            resp = self._send(req, author, content, ip, False)
            tree = ElementTree.fromstring(resp)
            karma = 0
            reason = []
            for entry in ('username', 'ip', 'email'):
                e = tree.find('./%s/appears' % entry)
                if e is not None and e.text == "1":
                    confidence = tree.find('./%s/confidence' % entry).text
                    karma += abs(self.karma_points) * float(confidence) / 100.0
                    reason.append("%s [%s]" % (entry, confidence))
            reason = ",".join(reason)
            if karma:
                karma = int(karma + 0.5)
                if karma < 1:
                    karma = 1;
                return (-karma,
                        N_("StopForumSpam says this is spam (%s)"), reason)
        except IOError, e:
            self.log.warn("StopForumSpam request failed (%s)", e)

    def train(self, req, author, content, ip, spam=True):
        if not spam:
            return 0
        elif not self._check_preconditions(True):
            return -2

        try:
            self._send(req, author, content, ip, True)
            return 1
        except urllib2.URLError, e:
            self.log.warn("StopForumSpam request failed (%s)", e)
        return -1

    # Internal methods

    def _check_preconditions(self, train):
        if self.karma_points == 0:
            return False

        if train and not self.api_key:
            return False

        return True

    def _send(self, req, author, content, ip, train):
        # Split up author into name and email, if possible
        author = author.encode('utf-8')
        author_name, author_email = parseaddr(author)
        if not author_name and not author_email:
            author_name = author
        elif not author_name and author_email.find('@') < 1:
            author_name = author
            author_email = None
        if author_name == 'anonymous':
            author_name = None

        params = {'ip': ip}
        if author_name:
            params['username'] = author_name
        if author_email:
            params['email'] = author_email

        if train:
            if not author_name or not author_email:
                return
            params['api_key'] = self.api_key
            params['ip_addr'] = ip
            params['evidence'] = "Spam training using Trac SpamFilter " \
                                 "(%s)\n%s" \
                                 % (user_agent, content.encode('utf-8'))
            url = 'http://www.stopforumspam.com/add.php'
            urlreq = urllib2.Request(url, urlencode(params),
                                     {'User-Agent': user_agent})
        else:
            params['ip'] = ip
            url = 'http://www.stopforumspam.com/api?confidence&f=xmldom&' + \
                  urlencode(params)
            urlreq = urllib2.Request(url, None,
                                     {'User-Agent': user_agent})

        resp = urllib2.urlopen(urlreq)
        return resp.read()



# -*- coding: utf-8 -*-
#
# Copyright (C) 2006-2016 Edgewall Software
# All rights reserved.
#
# This software is licensed as described in the file COPYING, which
# you should have received as part of this distribution. The terms
# are also available at https://trac.edgewall.com/license.html.
#
# This software consists of voluntary contributions made by many
# individuals. For the exact contribution history, see the revision
# history and logs, available at http://projects.edgewall.com/trac/.

from trac.db import Column, DatabaseManager, Table


def do_upgrade(env, ver, cursor):
    """Add a column to the log table for storing the request headers."""
    table = [
        Table('spamfilter_log', key='id')[
            Column('id', auto_increment=True),
            Column('time', type='int'),
            Column('path'),
            Column('author'),
            Column('authenticated', type='int'),
            Column('ipnr'),
            Column('headers'),
            Column('content'),
            Column('rejected', type='int'),
            Column('karma', type='int'),
            Column('reasons')
        ]
    ]
    DatabaseManager(env).upgrade_tables(table)
